import os

import torch
import torch.distributed as dist
from torch.nn.parallel import DistributedDataParallel as DDP


def check_distributed_setup() -> bool:
    """Tests distributed setup with a manual calculation."""
    if is_distributed():
        rank = dist.get_rank()
        world_size = dist.get_world_size()
        tensor = torch.arange(2, dtype=torch.int64) + 1 + 2 * rank
        dist.all_reduce(tensor, op=dist.ReduceOp.SUM)

        analytic_result = torch.tensor(
            [sum([0 + 1 + 2 * i for i in range(world_size)]), sum([1 + 1 + 2 * i for i in range(world_size)])]
        )

        return torch.allclose(tensor, analytic_result)
    return False


def is_same_accross_workers(local_tensor: torch.Tensor) -> bool:
    """Checks for a tensor if it is the same across all workers.
    Needs to be able to execute on all nodes, since we call all_gather. Otherwise this leads to deadlocks.
    """
    world_size = dist.get_world_size()

    destination_tensor_list = [torch.zeros_like(local_tensor) for _ in range(world_size)]
    dist.all_gather(destination_tensor_list, local_tensor)

    return all(torch.allclose(local_tensor, destination_tensor) for destination_tensor in destination_tensor_list)


def should_distribute() -> bool:
    """Returns true if job should be distributed."""
    return dist.is_available() and int(os.environ.get("WORLD_SIZE", 1)) > 1


def is_distributed() -> bool:
    """Returns true if job is distributed."""
    return dist.is_available() and dist.is_initialized()


def model_to_ddp(model: torch.nn.Module) -> DDP:
    """Converts a torch model into a distributed data parallel model."""
    model = torch.nn.SyncBatchNorm.convert_sync_batchnorm(model)
    # In the distributed setting, the devices always start at index 0 for every worker.
    # Currently we only support ddp, which uses one device per worker, so we always choose 0.
    return DDP(model.to("cuda:0"), device_ids=[0], output_device=0)



from contextlib import AbstractContextManager
from enum import Enum
from typing import List

MXM_THEME = {
    "title_font": "Space Grotesk",
    "text_font": "Inter",
    "primary": ["#000000", "#FFFFFF", "#381D59", "#FFFAF4"],
    "secondary": ["#ABABF9", "#FF5665", "#11D8C1", "#9E36FF"],
    "tertiary": ["#FF5665", "#9E36FF", "#CE3D4A", "#007D85"],
}


class MXM_STYLE(Enum):
    LIGHT = 1
    DARK = 2


def get_color_cycle(style: MXM_STYLE = MXM_STYLE.LIGHT) -> List[str]:
    """Get color cycle for the chosen style.

    Args:
        style (MXM_STYLE, optional): The style used for plotting. Defaults to MXM_STYLE.LIGHT.

    Returns:
        List[str]: List of colors for the color cycle.
    """
    if style == MXM_STYLE.LIGHT:
        return [
            MXM_THEME["tertiary"][1],
            MXM_THEME["secondary"][1],
            MXM_THEME["tertiary"][2],
            MXM_THEME["secondary"][3],
            MXM_THEME["tertiary"][3],
        ]
    else:
        return [
            MXM_THEME["secondary"][2],
            MXM_THEME["secondary"][1],
            MXM_THEME["tertiary"][0],
            MXM_THEME["primary"][3],
            MXM_THEME["secondary"][0],
        ]


def get_rc_params(style: MXM_STYLE = MXM_STYLE.LIGHT) -> dict:
    """Get rc (runtime configuration) parameters for the chosen style.
    Used for customizing the properties and default styles of Matplotlib.

    Some fonts are not available by default in Matplotlib, so you may need to install them, see :ref:`plotting`.

    Args:
        style (MXM_STYLE, optional): The style used for plotting. Defaults to MXM_STYLE.LIGHT.

    Returns:
        dict: Dictionary with rc parameters.
    """
    import matplotlib.pyplot as plt

    if style == MXM_STYLE.LIGHT:
        fg = MXM_THEME["primary"][0]
        bg = MXM_THEME["primary"][1]
    else:
        fg = MXM_THEME["primary"][3]
        bg = MXM_THEME["primary"][2]
    rc = {
        "axes.grid.axis": "y",
        "font.family": MXM_THEME["text_font"],
        "axes.prop_cycle": plt.cycler(color=get_color_cycle(style)),
        "savefig.transparent": True,
        "axes.spines.left": False,
        "axes.spines.right": False,
        "axes.spines.top": False,
        "axes.spines.bottom": False,
        "figure.facecolor": bg,
        "axes.facecolor": bg,
        "grid.color": fg,
        "text.color": fg,
        "axes.labelcolor": fg,
        "xtick.color": fg,
        "ytick.color": fg,
    }
    return rc


def get_mpl_context(style: MXM_STYLE = MXM_STYLE.LIGHT) -> AbstractContextManager[None]:
    """Get Matplotlib context manager for the chosen style.

    Args:
        style (MXM_STYLE, optional): The style used for plotting. Defaults to MXM_STYLE.LIGHT.

    Returns:
        AbstractContextManager[None]: Context manager for Matplotlib.
    """
    import matplotlib.pyplot as plt

    rc = get_rc_params(style)

    return plt.rc_context(rc=rc)



from typing import Optional

from omegaconf import MISSING

from scaffold.conf.scaffold.artifact_manager import ArtifactManagerConf
from scaffold.hydra.config_helpers import structured_config

GROUP = "scaffold/lightning/checkpointer"


@structured_config(group=GROUP)
class LightningCheckpointerConf:
    _target_: str = "scaffold.torch.lightning.callbacks.LightningCheckpointer"
    artifact_manager: ArtifactManagerConf = MISSING
    target_afid: Optional[str] = None  # artifact id to log the state of the model at the end of every epoch
    target_afid_best: Optional[str] = None  # artifact id to log the best model at the end of training
    resume_checkpoint_afid: Optional[str] = None  # artifact id to load a model state from at the start of training
    resume_checkpoint_version: Optional[int] = None  # If None, will load the latest version
    only_log_current_best: bool = False



import fsspec
from fsspec.core import split_protocol

from scaffold.data.constants import FILESYSTEM, URL

__all__ = ["get_fs_from_url", "get_protocol"]


def join_path(*parts: str) -> str:
    """Join path parts into a single path.

    This function concatenates the provided path segments using "/"
    and strips trailing slashes from each part. It works for both local
    paths and cloud bucket paths (e.g., "gs://").

    Args:
        *parts (str): Variable number of path segments.

    Returns:
        str: The resulting joined path.
    """
    return "/".join(part.rstrip("/") for part in parts if part)


def get_fs_from_url(url: URL, **storage_options) -> FILESYSTEM:
    """Get filesystem suitable for a url.

    Only the protocol part of the url is used, thus there is no need to call this method several times for different
    stores accessed with the same protocol. Should be called outside of store initialization to allow buffering from
    the same container among different stores or shards.

    The protocol can be overriden via the `storage_options`. This is important if people want to use the fsspec caching
    functionality, which requires the protocol to be, e.g., `simplecache`.
    """

    protocol, _ = split_protocol(url)

    # Provided storage_options take precedence
    storage_options = {"protocol": protocol, **storage_options}
    return fsspec.filesystem(**storage_options)


def get_protocol(url: str) -> str:
    """Get the protocol from a url, return empty string if local"""
    return f"{url.split('://')[0]}://" if "://" in url else ""



# Enables postponed evaluation of type hints. See https://www.python.org/dev/peps/pep-0563/
# This is especially useful when using type hints for modules that we want to import lazily.
from __future__ import annotations

import logging
from dataclasses import asdict
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import List, Optional

import torch
from pytorch_lightning import LightningModule, Trainer
from pytorch_lightning.callbacks import Callback
from pytorch_lightning.utilities import rank_zero_info, rank_zero_only

from scaffold.conf.scaffold.lightning.callbacks import LightningCheckpointerConf
from scaffold.data.artifact_manager.artifact_id_manager import validate_or_generate_target_afid
from scaffold.data.artifact_manager.artifact_logger import ModelLogger, STATE_FILENAME
from scaffold.data.artifact_manager.base import ArtifactManager
from scaffold.torch.distributed.ddp import is_distributed

logger = logging.getLogger(__name__)


class LightningCheckpointer(Callback):
    def __init__(
        self,
        artifact_manager: ArtifactManager,
        target_afid: Optional[str] = None,
        target_afid_best: Optional[str] = None,
        resume_checkpoint_afid: Optional[str] = None,
        resume_checkpoint_version: Optional[int] = None,
        only_log_current_best: bool = True,
    ):
        """Initializing the custom checkpointer.

        Notes:
            This method will create a local directory under /tmp (or equivalent depending on the OS) to save the
            best model checkpoint.

        Args:
            artifact_manager (ArtifactManager): ArtifactManager to use for logging.
            target_afid (str): Afid to log the state of the model and optimizers at the end of every epoch.
            target_afid_best (str): If set, will save the best model with this afid name. If None, will create a new
                random afid with the prefix 'best_model'
            resume_checkpoint_afid (str): Afid to load a model, optimizer and progress state from at the start of
                training.
            resume_checkpoint_version (int): Artifact version to load from. If None, will load the latest version.
            only_log_current_best (bool): If True, will only log a state, if the validation loss of an epoch
                is the lowest recorded one until then. If False, a state will be logged every epoch.
        """

        self.artifact_manager = artifact_manager
        self.model_logger = ModelLogger(artifact_manager=artifact_manager)
        self.only_log_current_best = only_log_current_best
        self.resume_checkpoint_afid = resume_checkpoint_afid
        self.resume_checkpoint_version = resume_checkpoint_version

        self.target_afid = validate_or_generate_target_afid(artifact_manager=artifact_manager, afid=target_afid)
        self.target_afid_best = validate_or_generate_target_afid(
            artifact_manager=artifact_manager, afid=target_afid_best, prefix="best_model"
        )
        self.best_state_dir = Path(TemporaryDirectory(prefix=f"{self.target_afid_best}__").name)
        self.best_state_path = self.best_state_dir / STATE_FILENAME

        self.lowest_avg_val_loss = None

    def _log_state_with_new_afid(
        self, model: torch.nn.Module, optimizers: List[torch.optim.Optimizer], current_epoch: int, **kwargs
    ) -> str:
        """Generates a new random state afid and logs it to the artifact store.

        Args:
            model (torch.nn.Module): Model to save the state off.
            optimizers (List[torch.optim.Optimizer]): All optimizers to save the state off.
            current_epoch (str): Current epoch.
            kwargs: Additional key value pairs to save to the state.
        """
        rank_zero_info(f"Saving state to afid {self.target_afid}")
        return self.model_logger.log_state_to_artifact(
            self.target_afid,
            model,
            optimizers,
            current_epoch=current_epoch,
            **kwargs,
        )

    @staticmethod
    def _get_avg_val_loss(callback_metrics: dict) -> float:
        try:
            return callback_metrics["avg_val_loss"]
        except KeyError as e:
            rank_zero_info("Please make sure to call self.log('avg_val_loss', avg_val_loss) in your lightning module.")
            raise e

    def on_train_start(self, trainer: Trainer, pl_module: LightningModule) -> None:
        """If an afid for a checkpoint was given, load the full state into the trainer and lightning module."""
        if self.resume_checkpoint_afid is not None:
            rank_zero_info(f"Loading state from afid {self.resume_checkpoint_afid}")
            state = self.model_logger.retrieve_state_from_artifact(
                afid=self.resume_checkpoint_afid, version=self.resume_checkpoint_version
            )
            # These attributes are read only on the trainer, so we have to assign to train_loop
            # See https://github.com/PyTorchLightning/pytorch-lightning/pull/7437/files for more info
            trainer.fit_loop.global_step = state.get("global_step", 0)
            trainer.fit_loop.current_epoch = state.get("current_epoch", 0)

            if trainer.max_epochs is not None and trainer.current_epoch > trainer.max_epochs:
                raise ValueError(
                    f"Trainer(max_epochs={trainer.max_epochs}),"
                    f" but the loaded checkpoint has current_epoch={trainer.current_epoch}!"
                )

            # Instead of overwriting the model object, we create and load the state dict.
            # This makes sure, that the model architecture did not change.
            if is_distributed():
                # In the distributed case, the model torch module is nested, so we need to assign it
                # so the previously saved state dict keys align properly.
                pl_module.model.module.load_state_dict(state["model"].state_dict())
            else:
                pl_module.model.load_state_dict(state["model"].state_dict())
            if "optimizer_state_dicts" in state:
                for optim, optim_state in zip(trainer.optimizers, state["optimizer_state_dicts"]):
                    optim.load_state_dict(optim_state)
            if "scheduler_state_dicts" in state:
                schedulers = [asdict(config) for config in trainer.strategy.lr_scheduler_configs]
                for scheduler, scheduler_state in zip(schedulers, state["scheduler_state_dicts"]):
                    for key in scheduler.keys():
                        # We need to make sure, that id(scheduler.optimzer) stays the same.
                        if key == "scheduler":
                            scheduler[key].load_state_dict(scheduler_state[key])
                        else:
                            scheduler[key] = scheduler_state[key]

    @rank_zero_only
    def on_validation_epoch_end(self, trainer: Trainer, pl_module: LightningModule) -> None:
        """Saving state dicts of models and optimizers after every validation run."""

        avg_val_loss = self._get_avg_val_loss(trainer.callback_metrics)
        schedulers = [asdict(config) for config in trainer.strategy.lr_scheduler_configs]
        log_state_kwargs = {
            "model": pl_module.model,
            "optimizers": trainer.optimizers,
            "schedulers": schedulers,
            "current_epoch": trainer.current_epoch,
            "global_step": trainer.global_step,
            "avg_val_loss": avg_val_loss,
        }
        if self.lowest_avg_val_loss is None or avg_val_loss < self.lowest_avg_val_loss:
            # We always log a copy, so we can save the best under self.best_state_dir
            self.model_logger.save_state(self.best_state_dir, **log_state_kwargs)
            if self.only_log_current_best:
                self._log_state_with_new_afid(**log_state_kwargs)
            self.lowest_avg_val_loss = avg_val_loss

        if not self.only_log_current_best:
            self._log_state_with_new_afid(**log_state_kwargs)

    @rank_zero_only
    def on_train_end(self, trainer: Trainer, pl_module: LightningModule) -> None:
        """Log the best state dicts under the target afid"""
        # It's possible that training end before anything was saved
        if self.best_state_path.is_file():
            rank_zero_info(f"Saving best state dicts to target_afid_best {self.target_afid_best}")
            self.artifact_manager.log_files(
                artifact_name=self.target_afid_best,
                local_path=self.best_state_dir,
            )

    @rank_zero_only
    def load_best_state(self) -> dict:
        """Convenience function to load the current local best state."""
        if self.best_state_path.is_file():
            # torch.load can lead to deadlocks when trying to load a state that contains a distributed model.
            # This is the case when torch.load() is not called on all ranks.
            # That is why we make sure to only save non distributed model objects.
            return torch.load(self.best_state_path, weights_only=False)

    @classmethod
    def from_config(cls, cfg: LightningCheckpointerConf) -> "LightningCheckpointer":
        """Instantiate a LightningCheckpointer from a configuration.

        Args:
            cfg: Configuration for the LightningCheckpointer.
        """
        from hydra.utils import instantiate

        artifact_manager = instantiate(cfg.artifact_manager)
        return cls(
            artifact_manager=artifact_manager,
            target_afid=cfg.target_afid,
            target_afid_best=cfg.target_afid_best,
            resume_checkpoint_afid=cfg.resume_checkpoint_afid,
            resume_checkpoint_version=cfg.resume_checkpoint_version,
            only_log_current_best=cfg.only_log_current_best,
        )



import logging
import os
import typing as t
from contextlib import AbstractContextManager, contextmanager, ExitStack

from omegaconf import DictConfig

from scaffold.hydra import compose

logger = logging.getLogger(__name__)


class LoggingContext(AbstractContextManager):
    """
    Context manager for configuring the logging system.
    It wraps the hydra's logging configuration and allows to specify verbosity level per module,
        see hydra documentation for more details: https://hydra.cc/docs/tutorials/basic/running_your_app/logging/

    Example usage:
    ```
        cfg = LoggingContext.DEFAULT_CONFIGURATION
        with LoggingContext(cfg) as log_context:
            logging.info("This message should be logged with the specified configuration.")
        logging.info("This error should be logged with default configuration.")
    ```
    """

    DEFAULT_CONFIGURATION = compose("scaffold/entrypoint/logging/default.yaml")
    SILENT_CONFIGURATION = compose("scaffold/entrypoint/logging/disabled.yaml")

    def __init__(
        self,
        log_config: DictConfig = DEFAULT_CONFIGURATION,
        verbose: t.Union[bool, str, t.Sequence[str]] = False,
    ):
        """
        Initialize the configuration of the logging context.
        """
        self.log_config = log_config
        self.verbose = verbose

    def __enter__(self):
        """
        Configure the logging system with the specified settings on entering the context.
        """
        from hydra.core.utils import configure_log

        configure_log(self.log_config, self.verbose)

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Exit the context (and ideally restore previous logging configuration).
        """
        # TODO we want to reset the logging configuration here, but there is no way to access the previous configuration
        # for now we ignore this problem, but mind that even after the context is exited, the logging configuration will
        # still be the same as the one set in this context
        pass


class TimerContext(AbstractContextManager):
    def __init__(self, module_name: str = None):
        """
        Initialize the timer context manager.

        Args:
            module_name (str): Optional name of the module to be displayed in the log message.
        """
        self.start_time = None
        self.module_name = module_name

    @staticmethod
    def _get_time_str(t1: int, t2: int) -> str:
        """Returns a human-readable string representation of the time delta between t1 and t2."""
        delta_sec = t2 - t1
        m, s = divmod(delta_sec, 60)
        h, m = divmod(m, 60)
        res = ""
        if h > 0:
            res += f"{h:.0f} hours "
        if m > 0:
            res += f"{m:.0f} min "
        res += f"{s:.2f} sec"
        return res

    def __enter__(self):
        """
        Enter the timer context and start the timer
        """
        from time import perf_counter

        self.start_time = perf_counter()

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Exit the timer context and log the time taken for execution
        """
        from time import perf_counter

        module_name = self.module_name or "Module"
        end_time = perf_counter()
        logger.info(f"{module_name} executed in {self._get_time_str(self.start_time, end_time)}")


class WandBContext(AbstractContextManager):
    """
    Context manager for configuring wandb, starting and exiting a wandb run.

    Example usage:
    ```
        with WandBContext(project="chameleon", entity="mg515") as wandb_ctx:
            # Do something with the WandB context
            wandb.log({"metric": 0.5})
    ```
    """

    def __init__(
        self,
        user: t.Optional[str] = None,
        base_url: str = None,
        entity: str = "mxm",
        run_id: t.Optional[str] = None,
        resume: bool = False,
        run_config: t.Optional[DictConfig] = None,
        **kwargs,
    ) -> None:
        """WandB Context that connects to our WandB instance, creates a WandB run and logs config to WandB instance.

        Args:
            user: Optional cloud username to retrieve wandb login secrets.
                  See Also: :func:`scaffold.wandb.helpers.wandb_environment_setup()`
            base_url: Base URL of WandB instance.
            entity: Name of the WandB entity.
            run_id: Flyte or WandB run ID.
            resume: Whether to resume previous run specified by run_id.
            run_config: Config passed to wandb init.
            **kwargs: Additional kwargs passed to wandb init. See https://docs.wandb.ai/ref/python/init.
        """

        from scaffold.wandb.helpers import wandb_environment_setup

        wandb_environment_setup(user)
        os.environ["WANDB_BASE_URL"] = base_url

        self.entity = entity
        self.kwargs = kwargs
        self.run = None
        self.run_config = run_config or DictConfig({})

        # If we are in flyte context and resume, we set the run_id to the current execution id
        self.run_id = os.environ.get("FLYTE_INTERNAL_EXECUTION_ID", run_id) if resume else run_id
        if resume and self.run_id is None:
            raise ValueError("Resume is set to True but run_id is None.")

        self.kwargs["resume"] = "allow" if resume else None
        logger.info(f"Run will be logged under wandb id: {self.run_id}")

    def __enter__(self):
        """Enters the context manager and starts a new WandB run."""
        import wandb

        if wandb.run is not None:
            logger.info(f"Using existing WandB run {self.run_id}.")
            self.run = wandb.run

        else:
            import flatten_dict

            config = flatten_dict.flatten(dict(self.run_config), reducer="dot")
            self.run = wandb.init(
                entity=self.entity,
                id=self.run_id,
                config=config,
                **self.kwargs,
            )
            logger.info(f"Started WandB run {self.run_id}.")
        return wandb

    def __exit__(
        self, exc_type: t.Optional[type], exc_val: t.Optional[Exception], exc_tb: t.Optional[t.Any]
    ) -> t.Optional[bool]:
        import wandb

        logger.info(f"Finishing WandB run {self.run_id}.")
        wandb.finish(exit_code=1 if exc_type is not None else 0)


@contextmanager
def combined_context(*contexts: t.List[AbstractContextManager]):
    """
    Combine multiple context managers into a stack, since contexts should be exited in reverse order
    in which they are entered. If an exception occurred, this order matters, as any context manager
    could suppress the exception, at which point the remaining managers will not even get notified

    Example use case:

    ```
    with combined_context(LoggingContext(), WandBContext()) as (logger, wandb):
        ...
    ```
    """
    with ExitStack() as stack:
        yield [stack.enter_context(cls) for cls in contexts]



"""
This file is strongly informed by the design of WandB's client published under MIT license

A good overview and explanation of the different logged quantities can be found
at: https://lambdalabs.com/blog/weights-and-bias-gpu-cpu-utilization/
"""

import abc
import threading
import time
from typing import Dict, List, Optional

import psutil
import pynvml

GPUHandle = object
SamplerDict = Dict[str, List[float]]
StatsDict = Dict[str, float]


class AbstractBackendSender(metaclass=abc.ABCMeta):
    """
    Abstract interface that works in conjunction with the system monitor.
    Should be used to configure where the system metrics are being sent to.
    """

    @abc.abstractmethod
    def publish(self, event: StatsDict, **kwargs) -> None:
        """Implements sending `event` to the corresponding backend."""
        pass


class SystemMonitor(object):
    _pid: int
    gpu_count: int
    _thread: Optional[threading.Thread]
    sampler: StatsDict
    network_init: Optional[Dict]
    _backend_sender: AbstractBackendSender
    _sample_rate_seconds: int
    _rank_prefix: str
    _gpu_key: str

    """
    Important Note:
        Since psutil is giving us system level information and not pod level information
        it is tricky to get reliable and not redundant CPU info here.
        We basically rely on the assumption that our rank is running on its own node here.
        With kubernetes this could be not true in a lot of scenarios, for example when one rank
        only requests one GPU, but one node in the training pool has more than one GPU.
        In this case, psutil will give node information, which could be used by other processes.
        One solution for this could be, to only allow the same number of GPUs per node as a worker
        would request, so that every worker has to request a new node.

        For GPU this is less of a problem, if the worker requests any number of GPUs and those GPUs are not
        shared by other workers / pods. In this case, all GPUs allocated to our worker pod will be the
        only visible cuda devices, and the GPU stats can only be attributed to our worker process.
    """

    def __init__(
        self,
        pid: int,
        backend_sender: AbstractBackendSender,
        sample_rate_seconds: int = 2,
        rank: Optional[int] = None,
    ) -> None:
        """
        Initializes the SystemMonitor using the parent threads PID and a backend_sender that implements the
        logic to send the collectesd stats to the corresponding backend, e.g. CSV, Tensorboard, WandB.

        Args:
            pid (int): Process id of the parent process.
            backend_sender (AbstractBackendSender): Class that publishes the stats, e.g. to WandB.
            sample_rate_seconds (int): The rate of sampling system stats in seconds.
            rank (int): Rank of the current process. Only needed in a distributed setting.
        """
        try:
            pynvml.nvmlInit()
            self.gpu_count = pynvml.nvmlDeviceGetCount()
        except pynvml.NVMLError:
            self.gpu_count = 0

        self._pid = pid
        self._backend_sender = backend_sender
        self._sample_rate_seconds = sample_rate_seconds
        self._shutdown = False
        self._thread = None
        self.sampler = {}
        self._rank_prefix = f"rank{rank}." if rank is not None else ""
        self._gpu_key = (self._rank_prefix + "gpu.{}.{}").format

        if psutil:
            net = psutil.net_io_counters()
            self.network_init = {"sent": net.bytes_sent, "recv": net.bytes_recv}

    def start(self) -> None:
        """Starts the the monitor by launching a separate thread."""
        if self._thread is None:
            self._shutdown = False
            self._thread = threading.Thread(target=self._thread_body)
        if not self._thread.is_alive():
            self._thread.start()

    @property
    def proc(self) -> psutil.Process:
        """Returns the process corresponding to the parents PID"""
        return psutil.Process(pid=self._pid)

    def _thread_body(self) -> None:
        while True:
            stats = self._stats()
            # Not all stats are refreshed at the same time. To circumvent ever-changing keys in the sampler dict
            # we iterate through the collected stats and see if the stat is on the sampler dict. If not we add a None
            # and then add it with the value. If yes, we instantiate the sample with the same value again to carry it
            # over. In this way we get the full sampler dict after some time. This method will become especially
            # important if we start averaging quantities over a longer aggregation window.
            for stat, value in stats.items():
                self.sampler[stat] = self.sampler.get(stat, None)
                self.sampler[stat] = value

            self.flush()
            if self._shutdown:
                return

            seconds = 0.0
            while seconds < self._sample_rate_seconds:
                time.sleep(0.1)
                seconds += 0.1
                if self._shutdown:
                    return

    def shutdown(self) -> None:
        """Closes the SystemMonitor's thread."""
        self.flush()
        self._shutdown = True
        try:
            if self._thread is not None:
                self._thread.join()
        finally:
            self._thread = None

    def flush(self) -> None:
        """Used the backend senders publish method to sent the latest metrics to the configured backend"""
        self._backend_sender.publish(self.sampler)

    def _stats(self) -> StatsDict:
        stats: StatsDict = {}
        stats.update(self._collect_gpu_stats())
        stats.update(self._collect_system_stats())

        return stats

    def _collect_system_stats(self) -> StatsDict:
        stats: StatsDict = {}
        # See the notes in the class docstring under which assumption the rank prefix is sensible for CPU stats.
        pre = self._rank_prefix
        if psutil:
            net = psutil.net_io_counters()
            sysmem = psutil.virtual_memory()
            stats[pre + "cpu"] = psutil.cpu_percent()
            stats[pre + "memory"] = sysmem.percent
            stats[pre + "network"] = {
                "sent": net.bytes_sent - self.network_init["sent"],
                "recv": net.bytes_recv - self.network_init["recv"],
            }
            stats[pre + "disk"] = psutil.disk_usage("/").percent
            stats[pre + "proc.memory.availableMB"] = sysmem.available / 1048576.0
            try:
                stats[pre + "proc.memory.rssMB"] = self.proc.memory_info().rss / 1048576.0
                stats[pre + "proc.memory.percent"] = self.proc.memory_percent()
                stats[pre + "proc.cpu.threads"] = self.proc.num_threads()
            except psutil.NoSuchProcess:
                pass
        return stats

    def _collect_gpu_stats(self) -> StatsDict:
        stats: StatsDict = {}
        for i in range(0, self.gpu_count):
            handle = pynvml.nvmlDeviceGetHandleByIndex(i)
            try:
                stats.update(self._get_gpu_utilization(i, handle))
            except pynvml.NVMLError:
                pass

            # Try to add power usage metrics separately. Some GPU's don't provide these infos and hence
            # we should catch them in a separate block as otherwise we'd not be recording any metrics.
            try:
                stats.update(self._get_gpu_power_usage(i, handle))
            except pynvml.NVMLError:
                pass

        return stats

    def _get_gpu_power_usage(self, i: int, handle: GPUHandle) -> StatsDict:
        stats: StatsDict = {}
        power_watts = pynvml.nvmlDeviceGetPowerUsage(handle) / 1000.0
        power_capacity_watts = pynvml.nvmlDeviceGetEnforcedPowerLimit(handle) / 1000.0
        power_usage = (power_watts / power_capacity_watts) * 100

        stats[self._gpu_key(i, "powerWatts")] = power_watts
        stats[self._gpu_key(i, "powerPercent")] = power_usage

        return stats

    def _get_gpu_utilization(self, i: int, handle: GPUHandle) -> StatsDict:
        stats: StatsDict = {}
        utilz = pynvml.nvmlDeviceGetUtilizationRates(handle)
        memory = pynvml.nvmlDeviceGetMemoryInfo(handle)
        temp = pynvml.nvmlDeviceGetTemperature(handle, pynvml.NVML_TEMPERATURE_GPU)

        stats[self._gpu_key(i, "gpu")] = utilz.gpu
        stats[self._gpu_key(i, "memory")] = utilz.memory
        stats[self._gpu_key(i, "memoryAllocated")] = (memory.used / float(memory.total)) * 100
        stats[self._gpu_key(i, "temp")] = temp

        return stats



import importlib.util


def register_all() -> None:
    """All @structured_config classes register themselves when imported / defined"""
    from scaffold.conf.scaffold import artifact_manager  # noqa:F401
    from scaffold.conf.scaffold import entrypoint  # noqa:F401

    if importlib.util.find_spec("flytekit") is not None:
        # only import in case we installed flyte extra package
        from scaffold.conf.scaffold import flyte_launcher  # noqa:F401



from __future__ import annotations

import copy
import gzip
import importlib
import logging
import os
import pathlib
import random
import string
import sys
import tarfile
import tempfile
from pathlib import Path
from typing import Dict, List, Optional, Sequence, Tuple, TYPE_CHECKING

import pyprojroot
from flytekit import Email, Slack
from flytekit.core.notification import Notification
from flytekit.exceptions.user import FlyteEntityNotExistException, FlyteInvalidInputException

try:
    # not available in older flyte versions
    from flytekit.exceptions.system import FlyteSystemUnavailableException

    imported_unavailable_exceptions = True
except ImportError:
    imported_unavailable_exceptions = False
from hydra.core.utils import configure_log, filter_overrides, JobReturn, run_job, setup_globals
from hydra.plugins.launcher import Launcher
from hydra.types import HydraContext, TaskFunction
from omegaconf import DictConfig, OmegaConf, open_dict

from scaffold.conf.scaffold.flyte_launcher import (
    ExecutionEnvironmentEnum,
    FlyteDockerImageConfig,
    FlyteNotificationConfig,
    FlyteNotificationEnum,
    FlyteWorkflowConfig,
)

if TYPE_CHECKING:
    from flytekit import LaunchPlan, Workflow
    from flytekit.core.base_task import PythonTask
    from flytekit.core.workflow import WorkflowBase
    from flytekit.remote.entities import FlyteLaunchPlan
    from flytekit.remote.remote import FlyteRemote

logger = logging.getLogger(__name__)


class FlyteLauncher(Launcher):
    def __init__(
        self,
        execution_environment: str,
        endpoint: str,
        build_images: bool,
        fast_serialization: bool,
        run: bool,
        workflow: FlyteWorkflowConfig,
        notifications: Optional[List[FlyteNotificationConfig]],
    ) -> None:
        """
        Construct Flyte launcher.

        Args:
            execution_environment (str): Can be either 'local' or 'remote'
            endpoint (str): The Flyte platform endpoint to connect to
            build_images (bool): whether the launcher should build the docker images containing the workflow code
            fast_serialization (bool): whether to use fast serialization to inject code into existing containers
            run (bool): whether the workflow is executed after registration
            workflow: (FlyteWorkflowConfig): the config related to workflow execution/registration
            notifications (Optional[List[FlyteNotificationConfig]]): A list of `FlyteNotificationConfig` objects,
                each specifying the notification settings for the workflow. These configurations define:
                    - **Phases**: Workflow execution phases that will trigger the notification.
                    - **Recipients**: A list of recipient identifiers (e.g., email addresses or Slack email).
                    - **Type**: The communication method for the notification, such as email or Slack.
        """
        self.config: Optional[DictConfig] = None
        self.hydra_context: Optional[HydraContext] = None
        self.workflow_config: FlyteWorkflowConfig = workflow

        assert execution_environment in [
            "local",
            "remote",
        ], f"Execution environment must be either 'local' or 'remote', but was '{execution_environment}'."
        self.execution_environment = execution_environment
        self.notifications = notifications
        self.endpoint = endpoint
        self.build_images = build_images
        self.fast_serialization = fast_serialization
        self.run = run

    def setup(
        self,
        *,
        hydra_context: HydraContext,
        task_function: TaskFunction,
        config: DictConfig,
    ) -> None:
        """Setup hydra config and context."""
        self.config = config
        self.hydra_context = hydra_context
        self.task_function = task_function

    def _hydra_launch_setup(self) -> None:
        setup_globals()
        assert self.config is not None
        assert self.hydra_context is not None
        configure_log(self.config.hydra.hydra_logging, self.config.hydra.verbose)
        sweep_dir = Path(str(self.config.hydra.sweep.dir))
        sweep_dir.mkdir(parents=True, exist_ok=True)

        logger.info("Launch job with Flyte Launcher.")
        logger.info(f"Sweep output dir : {sweep_dir}")

    def _extract_module_name_from_task_function(self) -> str:
        """Extract the module name from the @hydra.main decorated task function in the workflow file.
        The module of the task function at runtime should be __main__ since it is executed with
        `python path/to/workflow.py`.
        This function tries to extract the corresponding module name as 'path.to.workflow'.

        This is important, because flyte needs to find the workflow with the module notation, since __main__ is not
        defined inside the container when running remotely.
        """
        path_flag = False
        if os.getcwd() not in sys.path:
            sys.path.insert(0, os.getcwd())  # The src directory needs to be in the PATH in order to find module_name.
            path_flag = True

        main_module = sys.modules[self.task_function.__module__].__file__
        relative_path = os.path.relpath(main_module, os.getcwd())
        module_string = os.path.splitext(relative_path)[0].replace(os.sep, ".")
        if module_string.startswith("src."):
            # we expect a src dir to be in pythonpath, but not to be a valid python package, i.e. has an init file
            # therefore the flyte task can run into issues when trying to import package 'src'. So we dont want to
            # have module names starting with src
            module_string = module_string[4:]

        # Checking if the module if importable, without importing it again, since it is already imported as __main__.
        if importlib.util.find_spec(module_string) is None:
            raise ModuleNotFoundError(
                f"Could not find the module '{module_string}'.\n"
                "This module is assumed to be the module that contains your @hydra.main decorated function.\n"
                f"Please check if the path to your task function is a valid python module: {main_module.__file__}."
            )
        if path_flag:
            sys.path.remove(os.getcwd())

        return module_string

    @staticmethod
    def _get_workflows_and_tasks(module_string: str) -> Tuple[WorkflowBase, List[WorkflowBase], List[PythonTask]]:
        """Recursively find all PythonTask and WorkflowBase objects in the given module.
        We assume one workflow to be the "main_workflow" which ties it all together.
        Please refer to the docs of 'identify_main_workflow' for more details.

        Args:
            module_string (str): String like 'flyte_workflows.my_workflow' corresponding to a
                file like 'flyte_workflows/my_workflow.py'.
        """

        from flytekit.core.base_task import PythonTask
        from flytekit.core.python_function_task import PythonFunctionTask
        from flytekit.core.workflow import WorkflowBase

        from scaffold.flyte.core import identify_main_workflow

        main_workflow, flyte_entities = identify_main_workflow(module_string)

        # order of registration matters because flyte checks dependency relationships
        # tasks get separated and registered first
        # workflows are registered in reverse order of discovery (lowest recursion first)
        for entity in flyte_entities:
            if not isinstance(entity, PythonTask) and not isinstance(entity, WorkflowBase):
                raise TypeError(
                    f"Expected only entities of type PythonTask or WorkflowBase, but found '{type(entity)}' instead!"
                )
        tasks = [
            obj
            for obj in flyte_entities
            if isinstance(obj, PythonTask)
            and getattr(obj, "execution_mode", None) != PythonFunctionTask.ExecutionBehavior.DYNAMIC
        ]
        workflows = [obj for obj in reversed(flyte_entities) if isinstance(obj, WorkflowBase)]

        return main_workflow, workflows, tasks

    def _resolve_pipeline_version(self) -> str:
        """Generate a unique version string that can be used for tagging build images and the flyte pipeline.
        NOTE: Make sure that you set a unique pipeline version if you set it manually in the config.

        We try to generate a version based on this order of trials:
        1. Manually set version in config.
        2. Git repository hash + '-dirty' if uncommited changes exist + random string if flyte domain is `development`
        3. Random string as a fallback.

        Returns:
            str - the pipeline version
        """
        from git.exc import InvalidGitRepositoryError

        from scaffold.flyte.git import get_branch_identifier

        if (version := self.config.hydra.launcher.workflow.version) is not None:
            pipeline_version = version
            logger.info(f"Using specified pipeline version from config: '{pipeline_version}'.")
        else:
            try:
                pipeline_version = get_branch_identifier(
                    random_suffix=(self.config.hydra.launcher.workflow.domain == "development")
                )
                logger.info(f"Generated pipeline version based on git repo: '{pipeline_version}'.")
            except InvalidGitRepositoryError:
                pipeline_version = "".join(random.choices(string.ascii_uppercase + string.digits, k=10))
                logger.info(
                    "Could not find a git repository."
                    f" Generated pipeline version based on random string: '{pipeline_version}'."
                )
                if not self.build_images:
                    logging.info(
                        "No images will be build and no docker images with the random pipeline version exist.\n"
                        "Will use provided target versions to execute tasks in."
                    )
                    for image in [
                        self.config.hydra.launcher.workflow.default_image
                    ] + self.config.hydra.launcher.workflow.extra_images:
                        if image.target_image_version is None:
                            raise ValueError(
                                f"Version for image '{image.flyte_image_name}' ({image.target_image})"
                                f" was not explicitly set."
                            )
                        logging.info(
                            f"Setting image '{image.flyte_image_name}' ({image.target_image}) version to tag "
                            f"'{image.target_image_version}'."
                        )

        # Set config entry so that it can be interpolated to individual image descriptions
        self.config.hydra.launcher.workflow.version = pipeline_version
        return pipeline_version

    @staticmethod
    def _resolve_docker_context_to_root_project_dir(image_config: FlyteDockerImageConfig) -> Tuple[str]:
        """
        Resolve the docker context relative to the root of the `project.
        This enables the user to run the launcher from any subdirectory of the project.

        Args:
            image_config (FlyteDockerImageConfig): Image to build

        Returns:
            Tuple[str, str] - The relative path to the project root and the relative path to the dockerfile

        """
        # https://github.com/chendaniely/pyprojroot/blob/main/src/pyprojroot/here.py
        project_root = pyprojroot.here()
        logger.info(f"Identified project root as: {project_root}")

        # get relative path from current working directory to project root
        project_root_rel = os.path.relpath(project_root, start=os.getcwd())
        docker_context = os.path.join(
            project_root_rel,
            image_config.docker_context,
        )
        dockerfile_path = os.path.join(docker_context, image_config.dockerfile_path)
        return docker_context, dockerfile_path

    def _run_image_build(self) -> Tuple[str, Dict[str, str]]:
        """
        Build all images specified in hydra config and return names

        Returns:
            Tuple[str, Dict[str, str]] - The name of the default image as well as a mapping from flyte image names to
                                         their full identifiers.
        """
        from scaffold.flyte.image_builder import build_image

        extra_images = {}

        self._check_unique_image_repo_conf()

        with open_dict(self.config.hydra.launcher.workflow.default_image.buildargs) as build_args:
            build_args["BASE_IMAGE"] = self.config.hydra.launcher.workflow.default_image.base_image
            build_args["BASE_VERSION"] = self.config.hydra.launcher.workflow.default_image.base_image_version

        if (
            self.config.hydra.launcher.workflow.default_image.base_image
            == self.config.hydra.launcher.workflow.default_image.target_image
        ):
            logger.error(
                f"Base image and target image for default image "
                f"'{self.config.hydra.launcher.workflow.default_image.flyte_image_name}' are identical: "
                f"{self.config.hydra.launcher.workflow.default_image.base_image}.\n"
                f" Please use a different target instead of mutating an existing image"
            )

        docker_context, dockerfile = self._resolve_docker_context_to_root_project_dir(
            self.config.hydra.launcher.workflow.default_image
        )
        default_image_tag = build_image(
            dockerfile=dockerfile,
            workdir_path=docker_context,
            image_repo=self.config.hydra.launcher.workflow.default_image.target_image,
            image_tag=self.config.hydra.launcher.workflow.default_image.target_image_version,
            push_image=True,
            buildargs=self.config.hydra.launcher.workflow.default_image.buildargs,
            secrets=self.config.hydra.launcher.workflow.default_image.secrets,
            **OmegaConf.to_container(self.config.hydra.launcher.workflow.default_image.docker_kwargs),
        )

        if self.config.hydra.launcher.workflow.extra_images is not None:
            for extra_image in self.config.hydra.launcher.workflow.extra_images:
                with open_dict(extra_image.buildargs) as build_args:
                    build_args["BASE_IMAGE"] = extra_image.base_image
                    build_args["BASE_VERSION"] = extra_image.base_image_version

                if extra_image.base_image == extra_image.target_image:
                    logger.error(
                        f"Base image and target image for extra image "
                        f"'{self.config.hydra.launcher.workflow.default_image.flyte_image_name}' are identical: "
                        f"{self.config.hydra.launcher.workflow.default_image.base_image}.\n"
                        f" Please use a different target instead of mutating an existing image"
                    )
                docker_context, dockerfile = self._resolve_docker_context_to_root_project_dir(extra_image)
                tag = build_image(
                    dockerfile=dockerfile,
                    workdir_path=docker_context,
                    image_repo=extra_image.target_image,
                    image_tag=extra_image.target_image_version,
                    push_image=True,
                    buildargs=extra_image.buildargs,
                    secrets=extra_image.secrets,
                    **OmegaConf.to_container(extra_image.docker_kwargs),
                )
                extra_images[extra_image.flyte_image_name] = tag

        return default_image_tag, extra_images

    def _check_unique_image_repo_conf(self) -> None:
        """
        Check that all configured image repositories are unique.
        If not, this leads to unexpected behavior, as only the latest built image will be used for all tasks.

        Raises:
            ValueError: If any of the configured image repositories are not unique.
        """
        if self.config.hydra.launcher.workflow.extra_images is not None:
            repos = [e_cfg.target_image for e_cfg in self.config.hydra.launcher.workflow.extra_images] + [
                self.config.hydra.launcher.workflow.default_image.target_image
            ]
            if len(repos) != len(set(repos)):
                raise ValueError(
                    "All specified image repositories must be unique. "
                    "Please check your flyte launcher config of default and extra images."
                )

    def _get_and_check_image_tags_without_building(self) -> Tuple[str, Dict[str, str]]:
        """
        Check the existence of provided image details and compile full list of identifiers for flyte

        Returns:
            Tuple[str, Dict[str, str]] - The name of the default image as well as a mapping from flyte image names to
                                         their full identifiers.
        """
        import docker

        docker_client = docker.from_env()

        def _check_if_image_exists(image: str) -> None:
            try:
                docker_client.images.get_registry_data(image)
            except docker.errors.NotFound:
                msg = f"Could not find image: {image}!\n"
                raise UserWarning(msg)

        extra_images = {}
        default_image_tag = (
            f"{self.config.hydra.launcher.workflow.default_image.base_image}:"
            f"{self.config.hydra.launcher.workflow.default_image.base_image_version}"
        )
        _check_if_image_exists(default_image_tag)
        if self.config.hydra.launcher.workflow.extra_images is not None:
            for extra_image in self.config.hydra.launcher.workflow.extra_images:
                extra_image_uri = f"{extra_image.target_image}:{extra_image.target_image_version}"
                _check_if_image_exists(extra_image_uri)
                extra_images[extra_image.flyte_image_name] = extra_image_uri

        return default_image_tag, extra_images

    @staticmethod
    def _format_notifications(notifications_config: Optional[List[FlyteNotificationConfig]]) -> List[Notification]:
        """
        Converts FlyteNotificationConfig to a list of Flyte notification objects.
        """
        notifications: List[Notification] = []

        for notification in notifications_config:
            if notification.type == FlyteNotificationEnum.email:
                notifications.append(
                    Email(
                        phases=notification.phases,
                        recipients_email=notification.recipients,
                    )
                )
            elif notification.type == FlyteNotificationEnum.slack:
                notifications.append(
                    Slack(
                        phases=notification.phases,
                        recipients_email=notification.recipients,
                    )
                )

        return notifications

    @staticmethod
    def _create_launchplan(
        workflow: Workflow,
        cfg: DictConfig,
        module_name: str,
        idx: int,
        config_name: str,
        notifications: List[Notification],
    ) -> LaunchPlan:
        """Create a launch plan with the given config.

        This is created to make the workflows executable from the UI. With the templated launch plans, the
        user can execute the workflows from the UI without the need of explicitly typing/giving any config.
        The launchplan will have the overriden configs as the default input.

        If `cfg` contains cronjob configuration, it is applied to the LaunchPlan.

        Args:
            cfg (DictConfig): hydra configuration of the workflow.
            workflow (Workflow): workflow to be executed with the prepared LaunchPlan.
            module_name (str): The name of the python module containing the workflow.
            idx (int): idx of the respective hydra override belonging to this `cfg` object.
            config_name (str): Name of the base hydra configuration file used to launch the workflow.
            notifications (List[Notification]): A list of `Notification` objects, where each object defines:
                - **Phases**: The workflow execution phases that will trigger the notification.
                - **Recipients**: A list of recipient identifiers to be notified.
                - **Type**: The notification channel, such as email or Slack.

        Returns:
            A LaunchPlan object

        """
        from croniter import croniter
        from flytekit import CronSchedule, LaunchPlan

        cfg_arg_name = "cfg"
        kickoff_time_arg_name = "kickoff_time"
        kwargs = {}

        if cfg_arg_name not in workflow.interface.inputs.keys():
            raise ValueError(f"The arguments of the main workflow must contain '{cfg_arg_name}'")

        if (cron_exp := cfg.hydra.launcher.workflow.cron_schedule) is not None:
            assert idx == 0, "Cronjobs can't be registered with multiple overrides!"
            assert croniter.is_valid(cron_exp), f"Invalid cron expression {cron_exp}."

            kwargs["schedule"] = CronSchedule(
                schedule=cron_exp,
                kickoff_time_input_arg=kickoff_time_arg_name,
            )

        # Remove the hydra section that is passed to the Flyte workflow.
        # WARNING: If the hydra section is passed, set "config.hydra.job.id" and "config.hydra.job.num".
        # The hydra section can contain interpolation entries, which expect these values to exist.
        cfg = copy.deepcopy(cfg)
        with open_dict(cfg):
            del cfg["hydra"]

        templated_lp = LaunchPlan.create(
            f"hydra_workflow_cfg_{module_name}_{config_name}_{idx}",
            workflow=workflow,
            notifications=notifications,
            default_inputs={
                cfg_arg_name: cfg,
            },
            **kwargs,
        )
        if not hasattr(templated_lp, "__name__"):
            # necessary for flyte version 1.14.0, see https://github.com/flyteorg/flyte/issues/6062
            templated_lp.__name__ = templated_lp.name

        return templated_lp

    @staticmethod
    def _activate_launch_plan(remote: FlyteRemote, launchplan: FlyteLaunchPlan) -> None:
        """Activate a cron scheduled launch plan.

        Args:
            remote (FlyteRemote): a FlyteRemote object to send the update request
            launchplan (FlyteLaunchPlan): a remote flyte launchplan with cron scheduler to activate
        """
        from flytekit.models.launch_plan import LaunchPlanState

        remote.client.update_launch_plan(
            id=launchplan.id,
            state=LaunchPlanState.ACTIVE,
        )

    def _launch_local(self, job_overrides: Sequence[Sequence[str]], initial_job_idx: int) -> Sequence[JobReturn]:
        """
        Execute workflow locally.

        Method borrowed from hydra's basic launcher:
        https://github.com/facebookresearch/hydra/blob/809718cdcd64f9cd930d26dea69f2660a6ffa833/hydra/_internal/core_plugins/basic_launcher.py#L51
        """
        logger.info("Workflow executed locally, flyte workflow config is ignored.")

        setup_globals()
        assert self.hydra_context is not None
        assert self.config is not None
        assert self.task_function is not None

        configure_log(self.config.hydra.hydra_logging, self.config.hydra.verbose)
        sweep_dir = self.config.hydra.sweep.dir
        Path(str(sweep_dir)).mkdir(parents=True, exist_ok=True)
        logger.info(f"Launching {len(job_overrides)} jobs locally")
        runs: List[JobReturn] = []
        for idx, overrides in enumerate(job_overrides):
            idx = initial_job_idx + idx
            lst = " ".join(filter_overrides(overrides))
            logger.info(f"\t#{idx} : {lst}")
            sweep_config = self.hydra_context.config_loader.load_sweep_config(self.config, list(overrides))
            with open_dict(sweep_config):
                sweep_config.hydra.job.id = idx
                sweep_config.hydra.job.num = idx
            ret = run_job(
                hydra_context=self.hydra_context,
                task_function=self.task_function,
                config=sweep_config,
                job_dir_key="hydra.sweep.dir",
                job_subdir_key="hydra.sweep.subdir",
            )
            runs.append(ret)
            configure_log(self.config.hydra.hydra_logging, self.config.hydra.verbose)
        return runs

    def _launch_remote(self, job_overrides: Sequence[Sequence[str]], initial_job_idx: int) -> Sequence[JobReturn]:
        """Register tasks and workflow with flyte and optionally execute workflow."""

        # NOTE: We need these lazy imports, since hydra tries to register this plugin, even if you don't
        # set the launcher to hydra/launcher=flyte. This means, that users who want to install scaffold,
        # but want to use hydra without flyte, will get a user warning for not being able to import flytekit.

        from flytekit.configuration import FastSerializationSettings
        from flytekit.exceptions.user import FlyteEntityAlreadyExistsException
        from flytekit.tools.fast_registration import compute_digest
        from flytekit.tools.script_mode import tar_strip_file_attributes

        from hydra_plugins.flyte_launcher_plugin._flyte_ignore import FlyteIgnore
        from scaffold.flyte.core import get_serialization_settings, temp_flyte_remote

        # process configs and local directory structure
        self._hydra_launch_setup()

        config_name = self.config.hydra.job.config_name
        module_string = self._extract_module_name_from_task_function()
        main_workflow, workflows, tasks = self._get_workflows_and_tasks(module_string)
        pipeline_version = self._resolve_pipeline_version()

        configs, templated_lps = [], []

        notifications = self._format_notifications(self.notifications)

        for idx, overrides in enumerate(job_overrides):
            if len(overrides) > 0:
                config = self.hydra_context.config_loader.load_sweep_config(self.config, overrides)
            else:
                config = self.config

            configs.append(config)
            templated_lps.append(
                self._create_launchplan(main_workflow, config, module_string, idx, config_name, notifications)
            )

        # prepare docker image (or verify provided tags exist)
        if self.build_images:
            logging.info("Building images ...")
            default_image_tag, extra_images = self._run_image_build()
        else:
            logging.info("Skip building images ...")
            default_image_tag, extra_images = self._get_and_check_image_tags_without_building()

        logging.info(f"Using default image: {default_image_tag}")
        logging.info(f"Using extra images: {extra_images}")

        with temp_flyte_remote(
            project=self.config.hydra.launcher.workflow.project,
            domain=self.config.hydra.launcher.workflow.domain,
            endpoint=self.endpoint,
        ) as remote:
            # if in fast-serialisation mode, prepare archive to inject into existing container - altered copy of:
            # https://github.com/flyteorg/flytekit/blob/5dd887cf5cbe817a35d682a32886f304a97fc910/flytekit/remote/remote.py#L769
            # https://github.com/flyteorg/flytekit/blob/5dd887cf5cbe817a35d682a32886f304a97fc910/flytekit/remote/remote.py#L695
            # https://github.com/flyteorg/flytekit/blob/5dd887cf5cbe817a35d682a32886f304a97fc910/flytekit/tools/fast_registration.py#L22

            fast_serialization_settings = None
            if not self.build_images and self.fast_serialization:
                logging.info(
                    "Registering workflows and tasks in fast serialization mode!\n"
                    "This means that we will copy your code into the respective containers without building!"
                )

                # build and upload tar to inject in existing container
                ignorer = FlyteIgnore(".", ignore_path=self.config.hydra.launcher.workflow.ignore)
                digest = compute_digest(pathlib.Path("."), ignorer.is_ignored)
                with tempfile.TemporaryDirectory() as tmp_dir:
                    archive_fname = os.path.join(tmp_dir, f"fast_{digest}.tar.gz")
                    tar_path = os.path.join(tmp_dir, "tmp.tar")
                    with tarfile.open(tar_path, "w") as tar:
                        tar.add(".", arcname="", filter=lambda x: ignorer.tar_filter(tar_strip_file_attributes(x)))
                    with gzip.GzipFile(filename=archive_fname, mode="wb", mtime=0) as gzipped:
                        with open(tar_path, "rb") as tar_file:
                            gzipped.write(tar_file.read())
                    # upload archive
                    _, native_url = remote.upload_file(pathlib.Path(archive_fname))

                # create serialisation description for registration of new task versions
                fast_serialization_settings = FastSerializationSettings(
                    enabled=True,
                    destination_dir=".",
                    distribution_location=native_url,
                )

            settings = get_serialization_settings(
                default_image=default_image_tag,
                extra_images=extra_images,
                fast_serialization_settings=fast_serialization_settings,
                domain=self.config.hydra.launcher.workflow.domain,
                project=self.config.hydra.launcher.workflow.project,
            )

            # register workflows and tasks
            try:
                for task in tasks:
                    remote.register_task(task, serialization_settings=settings, version=pipeline_version)
                for workflow in workflows:
                    remote.register_workflow(workflow, serialization_settings=settings, version=pipeline_version)
                remote.register_workflow(main_workflow, serialization_settings=settings, version=pipeline_version)
            except (FlyteInvalidInputException, FlyteEntityAlreadyExistsException):
                logger.info(f"Workflow already registered with version {pipeline_version}.")
            except Exception as e:
                if imported_unavailable_exceptions and isinstance(e, FlyteSystemUnavailableException):
                    raise FlyteSystemUnavailableException(
                        "Couldnt reach flyte! Check if you have port forwarded flyteadmin service, "
                        "see https://docs.scaffold.merantix-momentum.cloud/usage/hydra_flyte_launcher/full_example.html"
                    )
                else:
                    raise e

            else:
                logger.info("Workflows and tasks successfully registered.")

            # fetch or register launchplan
            for config, templated_lp in zip(configs, templated_lps):
                try:
                    remote_lp = remote.fetch_launch_plan(
                        name=templated_lp.name,
                        version=pipeline_version,
                        project=self.config.hydra.launcher.workflow.project,
                        domain=self.config.hydra.launcher.workflow.domain,
                    )
                except FlyteEntityNotExistException:
                    # we need to register a new launch plan as we couldnt fetch one
                    lp_ver = pipeline_version + "_lp"  # flyte 1.14+ doesnt like same version for lp and tasks
                    settings.version = lp_ver
                    remote_lp = remote.register_launch_plan(
                        templated_lp,
                        version=lp_ver,
                        project=self.config.hydra.launcher.workflow.project,
                        domain=self.config.hydra.launcher.workflow.domain,
                        serialization_settings=settings,
                    )

                # Leave the inputs empty here, as it was configured while registering the LaunchPlan.
                if self.run is True:
                    if config.hydra.launcher.workflow.cron_schedule is not None:
                        self._activate_launch_plan(
                            remote=remote,
                            launchplan=remote_lp,
                        )
                        logger.info(
                            f"Launchplan {remote_lp.name} with cron schedule is activated "
                            f"with version: {pipeline_version}."
                        )
                    else:
                        remote.execute(
                            remote_lp, wait=False, inputs=templated_lp.saved_inputs, type_hints={"cfg": DictConfig}
                        )
                        logger.info(f"Workflow {main_workflow.name} is executed with version: {pipeline_version}.")

            if self.run is False:
                lp_names = " ".join([f"'{lp.name}'" for lp in templated_lps])
                logger.info(f"Launchplans {lp_names} are registered with version: {pipeline_version}.")

        return []

    def launch(self, job_overrides: Sequence[Sequence[str]], initial_job_idx: int) -> Sequence[JobReturn]:
        """
        Launch the workflow either locally or register the workflow with flyte and optionally execute it.
        Having this functionality is useful for offering ability to switch between local and remote execution
        without changing the entire hydra config, but only a single flag.

        Args:
            job_overrides: a List of List<String>, where each inner list is the arguments for one job run.
            initial_job_idx: Initial job idx in batch.

        Returns:
            an array of return values from run_job with indexes corresponding to the input list indexes.
        """

        if self.execution_environment == ExecutionEnvironmentEnum.local:
            return self._launch_local(job_overrides, initial_job_idx)
        elif self.execution_environment == ExecutionEnvironmentEnum.remote:
            return self._launch_remote(job_overrides, initial_job_idx)
        else:
            raise ValueError(
                f"Execution environment must be either 'local' or 'remote', but was '{self.execution_environment}'."
            )



import logging
import os
import re
import shutil
import tempfile
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Iterable, Optional, Union

from scaffold.data.fs import join_path

logger = logging.getLogger(__name__)


class TmpArtifact:
    """Context manager for temporary artifact download.

    This class downloads an artifact to a temporary directory when entering
    a context, and cleans up the directory upon exit.
    """

    def __init__(self, artifact_manager: "ArtifactManager", collection: str, artifact: str, version: str) -> None:
        """Initialize a temporary artifact context.

        Args:
            artifact_manager (ArtifactManager): The artifact manager to use.
            collection (str): The collection name.
            artifact (str): The artifact name.
            version (str): The artifact version.
        """
        self.artifact_manager = artifact_manager
        self.tempdir = tempfile.mkdtemp()
        self.collection = collection
        self.artifact = artifact
        self.version = version

    def __enter__(self) -> str:
        """Download the artifact into a temporary directory.

        Returns:
            str: The path to the temporary directory containing the artifact.
        """
        self.artifact_manager.download_artifact(self.artifact, self.collection, self.version, to=self.tempdir)
        return self.tempdir

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        """Clean up the temporary directory."""
        shutil.rmtree(self.tempdir)


class DirectoryLogger:
    """Context manager for logging a directory as an artifact.

    This class creates a temporary directory where files can be written.
    When exiting the context, if the directory contains files, it is logged
    as an artifact.
    """

    def __init__(self, artifact_manager: "ArtifactManager", artifact: str, collection: Optional[str] = None) -> None:
        """Initialize a DirectoryLogger.

        Args:
            artifact_manager (ArtifactManager): The artifact manager to use.
            artifact (str): The artifact name.
            collection (Optional[str]): The collection name. Defaults to the artifact manager's active collection.
        """
        self.artifact_manager = artifact_manager
        self.artifact = artifact
        self.collection = collection or artifact_manager.active_collection
        self.tempdir = tempfile.mkdtemp()

    def __enter__(self) -> str:
        """Create and return a directory for logging files.

        Returns:
            str: The path to the logging directory.
        """
        self.artifact_dir = join_path(self.tempdir, self.artifact)
        os.makedirs(self.artifact_dir, exist_ok=False)
        return self.artifact_dir

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        """Log the folder if non-empty and clean up the temporary directory."""
        if os.listdir(self.artifact_dir):
            self.artifact_manager.log_files(self.artifact, self.artifact_dir, self.collection)
        shutil.rmtree(self.tempdir)


class ArtifactManager(ABC):
    def __init__(self, collection: str = "default"):
        """Artifact manager interface for various backends."""
        self._active_collection = collection

    @property
    def active_collection(self) -> str:
        """
        Collections act as folders of artifacts.

        It is ultimately up to the user how to structure their artifact store and the collections therein. All
        operations accessing artifacts allow explicit specification of the collection to use.
        Therefore, users could use collections to separate different artifact types or different experiments / runs.
        The manager maintains an 'active' collection which it logs to by default. This can be set at the run start
        when the manager is initialized and then remain unchanged for the duration of the run.

        To avoid incompatibility between different backends, collections cannot be nested (e.g. as subfolders on a
        filesystem) as in particular the WandB backend has no real notion of nested folder structures.
        """
        return self._active_collection

    @active_collection.setter
    def active_collection(self, value: str) -> None:
        """
        Sets the active collections that is being logged to by default.

        The provided value is verified to ensure that no nested collections are used.
        """
        if not re.match(r"^[a-zA-Z0-9\-_:]+$", value):
            raise ValueError(
                "Invalid collection name - must not be empty and can only contain alphanumerics, dashes, underscores "
                "and colons."
            )
        self._active_collection = value

    @abstractmethod
    def list_collection_names(self) -> Iterable:
        """Return list of all collections in the artifact store"""
        raise NotImplementedError

    @abstractmethod
    def exists_in_collection(self, artifact: str, collection: Optional[str] = None) -> bool:
        """Check if artifact exists in specified collection."""
        raise NotImplementedError

    @abstractmethod
    def log_files(
        self,
        artifact_name: str,
        local_path: Path,
        collection: Optional[str] = None,
        artifact_path: Optional[Path] = None,
    ) -> None:
        """
        Upload a file or folder into (current) collection, increment version automatically

        Args:
            artifact_name: Name of artifact to log
            local_path: Local path to the file or folder to log
            collection: Name of collection to log to, defaults to the active collection
            artifact_path: path under which to log the files within the artifact, defaults to "./"
        """
        raise NotImplementedError

    @abstractmethod
    def download_artifact(
        self, artifact: str, collection: Optional[str] = None, version: Optional[str] = None, to: Optional[str] = None
    ) -> Union[str, TmpArtifact]:
        """
        Download artifact contents (from current collection) to specific location and return a source listing them.

        If no target location is specified, a context manager for a temporary directory is created and the path to it
        is returned. Retrieve latest version unless specified.
        """
        raise NotImplementedError

    def exists(self, artifact: str) -> bool:
        """Check if artifact exists in specified collection."""
        return any([self.exists_in_collection(artifact, collection) for collection in self.list_collection_names()])

    def log_folder(self, artifact: str, collection: Optional[str] = None) -> DirectoryLogger:
        """Create a context manager for logging a directory of files as a single artifact."""
        return DirectoryLogger(self, artifact, collection)



import typing as t

from flytekit.configuration import Config as FlyteConfig
from flytekit.remote import FlyteLaunchPlan, FlyteRemote


class FlyteRemoteHelper:
    def __init__(self, domain: str, admin_endpoint: str, insecure: bool = True, project: str = "default"):
        """
        Instantiates a flyteRemote object internally.

        Args:
            domain: In which flyte domain we operate, usually one of staging, development, production
            project: flyte project to operate in. Defaults to 'default'
            admin_endpoint: Connection to Flyte admin. Usually 'flyteadmin.flyte.svc.cluster.local:81' within the
                cluster, or localhost:30081 if forwarded
                via `kubectl port-forward --address 0.0.0.0 svc/flyteadmin 30081:81 -n flyte`
            insecure: If true, no SSL is used for the connection. If the connection is just within cluster or through
                a port-forward done by kubectl it can be set to True.
        """
        self.flyte_remote = FlyteRemote(
            config=FlyteConfig.for_endpoint(endpoint=admin_endpoint, insecure=insecure),
            default_domain=domain,
            default_project=project,
        )

    def fetch_launchplan(self, launchplan_name: str) -> FlyteLaunchPlan:
        """Fetches a registered, remote launchplan by name."""
        return self.flyte_remote.fetch_launch_plan(name=launchplan_name)

    def execute_flyte_launchplan(self, launchplan_name: str, input_args: dict, wait: bool = False) -> t.Any:
        """
        Fetches a launchplan using the connection from init and executes it with the given arguments.

        Args:
            launchplan_name: Name of the launchplan.
                Note that launchplans registered via scaffold usually have a _0 suffix
            input_args: A dictionary mapping from worklow input arguments as string keys to the values with which the
                workflow gets executed. All arguments that do not have a default argument in the launchplan need to be
                provided. Note that cfg does have a default argument in launchplans registered using the flyte launcher.
            wait: If true, this function will only return once the workflow finishes. Defaults to false.

        Returns:
            something
        """
        lp = self.fetch_launchplan(launchplan_name)
        r = self.flyte_remote.execute(lp, inputs=input_args, wait=wait)
        return r



from dataclasses import field
from typing import Any, Dict, List, Optional

from omegaconf import MISSING

from scaffold.hydra.config_helpers import structured_config

GROUP = "scaffold/entrypoint"


@structured_config(group=GROUP)
class EntrypointConf:
    defaults: List[Any] = field(
        default_factory=lambda: [
            {"/scaffold/entrypoint/logging@logging": "default"},
        ]
    )
    # We unfortunately have to use Any inside 'contexts' instead of e.g. ContextConf,
    # because otherwise all context configs can't add additional arguments.
    # Hydra does not allow for Union types, which would be required, of multiple types of configs should be allowed.
    # Hydra itself solves it this way too until now: https://github.com/facebookresearch/hydra/blob/796fdf1c54eabdcd5121d3a9c8ab6af70ee41763/hydra/conf/__init__.py#L142  # noqa E501
    contexts: Optional[Dict[str, Any]] = field(default_factory=lambda: {})
    logging: Dict[str, Any] = MISSING  # Same options as hydra.job_logging
    verbose: Any = False  # Same as hydra.verbose, but applied to our logging setup, instead of hydra.job_logging



import warnings
from copy import deepcopy
from functools import reduce
from inspect import isclass
from typing import List, Optional, Union

import hydra
from hydra.core.config_store import ConfigStore
from omegaconf import DictConfig, OmegaConf

import scaffold
from scaffold.hydra.config_helpers import StructuredConfig

CONFIG_STORE = ConfigStore.instance()


def _compose_path(
    config_path: str,
    config_dir: Optional[str] = None,
    overrides: Optional[List[str]] = None,
    return_hydra_config: bool = False,
) -> DictConfig:
    """Initialized hydra if needed and composes a config from the config store, either yaml of structured config."""
    with scaffold.hydra.initialize(config_dir=config_dir):
        return hydra.compose(
            config_path.lstrip("/"), overrides=overrides or [], return_hydra_config=return_hydra_config
        )


def _instance_warning(config: StructuredConfig) -> None:
    """Trigger a warning if a StructuredConfig instance is passed."""
    if isinstance(config, StructuredConfig) and config != config.__class__():
        warnings.warn(
            "Instantiating a structured config with different argument has no effect on composition!"
            "The reason is that defaults list and dataclass fields behave differently during instantiation."
            f" Will compose {config.__class__.__name__} class with default values."
        )


def compose(
    config_name_or_class: Union[str, StructuredConfig],
    config_dir: Optional[str] = None,
    overrides: Optional[List[str]] = None,
    return_hydra_config: bool = False,
    check_missing: Optional[bool] = False,
    return_leaf: Optional[bool] = True,
) -> DictConfig:
    """Alternative for hydra.compose, which differs in the following ways:

    1. No need to call `hydra.initialize` before composing a config, since this will use `scaffold.hydra.initialize`
       internally, which uses an existing instance if available.
    2. Is able to compose a config from a `StructuredConfig` class, which does not require the user to know the path
       in the config store, where the structured config was registered.
    3. Can check for missing values right away, instead of throwing an exception when trying to access them.
    4. Can automatically return the leaf node of the given config. When calling the original
       hydra.compose("/my/grouped/config"), this results in a config with the keys config["my"]["group"][...].
       Setting return_leaf=True, scaffold.hydra.compose will automatically return the result of config["my"]["group"]
       instead of adding all group keys.

    Args:
        config_name_or_class (Union[str, StructuredConfig]): Path to the config in the config store, or StructuredConfig
            class. Passing an instance of StructuredConfig has the same effect, but will throw a warning if the instance
            arguments are not the same as the default ones.
        config_dir (Optional[str] = None): Optional absolute config directory to add to the search path.
        overrides (List[str] = None): Hydra overrides. See https://hydra.cc/docs/advanced/override_grammar/basic/.
        return_hydra_config (bool = False): If True, returns the hydra in the root of the config.
        check_missing (Optional[bool] = False): If True, will throw an exception if omegaconf.MISSING or '???' values
            are present.
        return_leaf (Optional[bool] = True): If True, will return the leaf node of the config path inside the config
            store, so compose("/my/grouped/config") will return only the values in the "config" node or yaml, without
            still containing the group keys in the returned config.

    Returns:
        Rendered DictConfig.
    """
    _config = deepcopy(config_name_or_class)
    compose_args = dict(config_dir=config_dir, overrides=overrides or [], return_hydra_config=return_hydra_config)

    if isinstance(_config, str):
        _config = _compose_path(_config, **compose_args)
    elif isinstance(config_name_or_class, StructuredConfig) or (
        isclass(_config)  # The separate isclass check is unfortunately needed.
        and issubclass(_config, StructuredConfig)
    ):
        _instance_warning(_config)
        store_path = _config.get_store_path()  # This is a class method, not an instance method
        _config = _compose_path(store_path, **compose_args)
    else:
        raise ValueError(f"Expected config to be of type Union[str, StructuredConfig], but got {type(_config)}.")

    if return_leaf:
        if isinstance(config_name_or_class, str):
            parent_keys = config_name_or_class.strip("/").split("/")[:-1]
        else:
            # Will result in an empty list, if the store path is just "MyConfig" and has not parents.
            parent_keys = config_name_or_class.get_store_path().split("/")[:-1]

        # Index through all levels (e.g. ["scaffold", "entrypoint", ...]) and returns the last result
        _config = reduce(DictConfig.get, parent_keys, _config)
        # Detach from original root, so that exceptions for missing values appear relative to the leaf.
        _config = DictConfig(OmegaConf.to_container(_config))

    if check_missing:
        # This triggers omegaconf.errors.MissingMandatoryValue for MISSING values.
        OmegaConf.to_object(_config)

    return _config



import os
import random
import string

from git import Repo


def get_branch_identifier(random_suffix: bool = False) -> str:
    """
    Get version from git repository.

    By default, the version is `branch-sha`.
    If there are untracked files, the suffix `-dirty` is appended.
    If `random_suffix=True` a random suffix is appended.

    Kwargs:
        random_suffix (bool): whether a random suffix should be appended.
    """
    repo = Repo(os.getcwd(), search_parent_directories=True)
    sha = repo.git.rev_parse(repo.head, short=True)

    if not repo.head.is_detached:
        version = repo.active_branch.name + "-" + sha
    elif (tag := next((tag for tag in repo.tags if tag.commit == repo.head.commit), None)) is not None:
        version = tag.name + "-" + sha
    else:
        version = repo.head.commit.hexsha

    # Make sure that the version prefix is a valid url path
    version = version.replace("/", "-")
    version = version.replace(".", "_")

    # untracked files or untracked changes or uncommitted but staged changes
    if (len(repo.untracked_files) > 0) or (len(repo.index.diff(None)) > 0) or (len(repo.index.diff("HEAD")) > 0):
        version += "-dirty"

    if random_suffix:
        suffix_len = 5
        version += "-" + "".join(random.choices(string.ascii_uppercase + string.digits, k=suffix_len))

    return version



import logging
import typing as t

import wandb

from scaffold.data.artifact_manager.base import ArtifactManager, TmpArtifact
from scaffold.data.fs import get_fs_from_url

logger = logging.getLogger(__name__)


class WandbArtifactManager(ArtifactManager):
    """Artifact manager that uses Weights & Biases (WandB) as the backend.

    This manager logs artifacts to WandB. In WandB, collections correspond to artifact types.
    Each artifact can contain multiple files, and versions are managed implicitly.

    Note:
        It is assumed that wandb.init() has been called before using this manager.
    """

    def __init__(
        self, entity: t.Optional[str] = None, project: t.Optional[str] = None, collection: str = "default"
    ) -> None:
        """Initialize the WandbArtifactManager.

        Args:
            entity (Optional[str]): If not provided, it's inferred from the active run / API settings.
            project (Optional[str]): If not provided, it's inferred from the active run / API settings.
            collection (str): The default collection (artifact type) name. Defaults to "default".

        Raises:
            ValueError: If neither a project nor an active project can be identified.
        """
        super().__init__(collection=collection)
        if project is not None:
            self.project = project
        elif wandb.run is not None:
            self.project = wandb.run.project
            logger.info(f"Using project {self.project} from active wandb run.")
        elif wandb.Api().settings["project"] is not None:
            self.project = wandb.Api().settings["project"]
            logger.info(f"Using project {self.project} from wandb settings.")
        else:
            raise ValueError("No project name was provided and no active project could be identified.")

        if entity is not None:
            self.entity = entity
        elif wandb.run is not None and wandb.run.entity is not None:
            self.entity = wandb.run.entity
            logger.info(f"Using entity {self.entity} from active wandb run.")
        elif wandb.Api().settings["entity"] is not None:
            self.entity = wandb.Api().settings["entity"]
            logger.info(f"Using entity {self.entity} from wandb settings.")
        else:
            self.entity = wandb.Api().project(self.project).entity
            logger.info(f"Using entity {self.entity} from wandb project.")

    def list_collection_names(self) -> t.Iterable[str]:
        """List all collections (artifact types) managed by WandB.

        Returns:
            Iterable[str]: A list of collection names corresponding to WandB artifact types.
        """
        return [collection.name for collection in wandb.Api().artifact_types(project=self.project)]

    def exists_in_collection(self, artifact: str, collection: t.Optional[str] = None) -> bool:
        """Check if an artifact exists in the specified collection.

        Note:
            The WandB API does not provide a direct existence check, so this is implemented
            by listing all artifacts in the collection.

        Args:
            artifact (str): The artifact name.
            collection (Optional[str]): The collection (artifact type) name. Defaults to the active collection.

        Returns:
            bool: True if the artifact exists in the collection, False otherwise.
        """
        collection = collection or self.active_collection
        if collection not in self.list_collection_names():
            return False
        return artifact in [
            art.name for art in wandb.Api().artifact_type(type_name=collection, project=self.project).collections()
        ]

    def log_files(
        self,
        artifact_name: str,
        local_path: str,
        collection: t.Optional[str] = None,
        artifact_path: t.Optional[str] = None,
    ) -> None:
        """Log files or a directory as a WandB artifact.

        This method uploads the file or directory located at `local_path` to WandB. If `local_path` is a directory,
        it uses `add_dir`; otherwise, it uses `add_file`.

        Args:
            artifact_name (str): The artifact name.
            local_path (str): The local path to the file or directory to be logged.
            collection (Optional[str]): The collection (artifact type) name. Defaults to the active collection.
            artifact_path (Optional[str]): An optional subpath within the artifact.
        """
        collection = collection or self.active_collection
        artifact = wandb.Artifact(artifact_name, type=collection)
        fs = get_fs_from_url(local_path)
        if fs.isdir(local_path):
            artifact.add_dir(local_path, name=artifact_path)
        else:
            artifact.add_file(str(local_path), name=artifact_path)
        artifact.save()
        artifact.wait()

    def download_artifact(
        self,
        artifact: str,
        collection: t.Optional[str] = None,
        version: t.Optional[str] = None,
        to: t.Optional[str] = None,
    ) -> t.Union[str, TmpArtifact]:
        """Download a specific artifact to a local path.

        WandB artifacts are downloaded in a nested folder structure.

        Args:
            artifact (str): The name of the artifact to download.
            collection (Optional[str]): The collection (artifact type) name. Defaults to the active collection.
            version (Optional[str]): The version of the artifact to download. If None, 'latest' is used.
            to (Optional[str]): The local destination path where the artifact should be downloaded.
                If not provided, a TmpArtifact context manager is returned.

        Returns:
            Union[str, TmpArtifact]: The local path where the artifact was downloaded, or a
                TmpArtifact context manager if no destination is provided.
        """
        collection = collection or self.active_collection
        if version is None:
            version = "latest"
        if wandb.run is None:
            art = wandb.Api().artifact(f"{self.entity}/{self.project}/{artifact}:{version}", type=collection)
        else:
            art = wandb.run.use_artifact(f"{self.entity}/{self.project}/{artifact}:{version}", type=collection)

        if to is not None:
            art.download(to)
            return to
        else:
            return TmpArtifact(self, collection, artifact, version)



import typing as t

from scaffold.data.artifact_manager.base import ArtifactManager, TmpArtifact
from scaffold.data.fs import get_fs_from_url, join_path


class FileSystemArtifactManager(ArtifactManager):
    """Artifact manager backed by a file system using fsspec.

    This implementation logs and retrieves artifacts from a specified URL.
    """

    def __init__(self, url: str, collection: str = "default", **fs_kwargs: t.Any) -> None:
        """Initialize a FileSystemArtifactManager.

        Args:
            url (str): The base URL of the artifact store.
            collection (str): The default collection name. Defaults to "default".
            **fs_kwargs (Any): Additional keyword arguments for the file system.
        """
        super().__init__(collection=collection)
        self.url = url.rstrip("/")
        self.fs = get_fs_from_url(self.url, **fs_kwargs)

    def list_collection_names(self) -> t.Iterable[str]:
        """List all collections in the artifact store.

        Returns:
            Iterable[str]: A list of collection names.
        """
        try:
            items = self.fs.ls(self.url, detail=True)
        except FileNotFoundError:
            return []
        return [item["name"].split("/")[-1] for item in items if item.get("type") == "directory"]

    def exists_in_collection(self, artifact: str, collection: t.Optional[str] = None) -> bool:
        """Check if an artifact exists in a specific collection.

        Args:
            artifact (str): The artifact name.
            collection (Optional[str]): The collection name. Defaults to the active collection.

        Returns:
            bool: True if the artifact exists, False otherwise.
        """
        collection = collection or self.active_collection
        artifact_dir = join_path(self.url, collection, artifact)
        return self.fs.exists(artifact_dir)

    def log_files(
        self,
        artifact_name: str,
        local_path: str,
        collection: t.Optional[str] = None,
        artifact_path: t.Optional[str] = None,
    ) -> None:
        """Log a file or folder as an artifact.

        This method uploads the file (or folder) located at `local_path` to the artifact store.
        If `artifact_path` is provided, the file is uploaded to a subpath within the artifact;
        otherwise, the entire folder is uploaded.

        Args:
            artifact_name (str): The artifact name.
            local_path (str): The local path to the file or folder.
            collection (Optional[str]): The collection name. Defaults to the active collection.
            artifact_path (Optional[str]): The subpath within the artifact for single file uploads.
        """
        collection = collection or self.active_collection
        base_artifact_path = join_path(self.url, collection, artifact_name)
        # Determine new version; start at "v0"
        if self.fs.exists(base_artifact_path):
            try:
                entries = self.fs.ls(base_artifact_path, detail=True)
                version_nums: t.List[int] = []
                for entry in entries:
                    ver = entry["name"].split("/")[-1]
                    if ver.startswith("v") and ver[1:].isdigit():
                        version_nums.append(int(ver[1:]))
                new_version = f"v{max(version_nums) + 1}" if version_nums else "v0"
            except Exception:
                new_version = "v0"
        else:
            new_version = "v0"
        target_dir = join_path(base_artifact_path, new_version)

        if artifact_path:
            # Upload a single file to target_dir/artifact_path.
            target_file = join_path(target_dir, artifact_path)
            fs = get_fs_from_url(target_file)
            parent_dir = fs._parent(target_file)
            self.fs.mkdirs(parent_dir, exist_ok=True)
            self.fs.put(local_path, target_file, recursive=False)
        else:
            # Upload an entire folder.
            if not self.fs.exists(target_dir):
                self.fs.mkdirs(target_dir, exist_ok=True)
            self.fs.put(local_path, target_dir, recursive=True)

    def download_artifact(
        self,
        artifact: str,
        collection: t.Optional[str] = None,
        version: t.Optional[str] = None,
        to: t.Optional[str] = None,
    ) -> t.Union[str, TmpArtifact]:
        """Download an artifact from the artifact store.

        If a destination path `to` is provided, the contents of the artifact version are copied
        there. Otherwise, a TmpArtifact context manager is returned.

        Args:
            artifact (str): The artifact name.
            collection (Optional[str]): The collection name. Defaults to the active collection.
            version (Optional[str]): The version of the artifact. If None or "latest", the latest version is used.
            to (Optional[str]): The destination path. If not provided, a temporary directory is used.

        Returns:
            Union[str, TmpArtifact]: The destination path or a TmpArtifact context manager.
        """
        collection = collection or self.active_collection
        base_artifact_path = join_path(self.url, collection, artifact)
        if version is None or version == "latest":
            if not self.fs.exists(base_artifact_path):
                raise ValueError(f"Artifact {artifact} not found in collection {collection}")
            entries = self.fs.ls(base_artifact_path, detail=True)
            version_nums: t.List[int] = []
            for entry in entries:
                ver = entry["name"].split("/")[-1]
                if ver.startswith("v") and ver[1:].isdigit():
                    version_nums.append(int(ver[1:]))
            if not version_nums:
                raise ValueError(f"No version found for artifact {artifact} in collection {collection}")
            version = f"v{max(version_nums)}"
        remote_artifact_path = join_path(base_artifact_path, version)
        if to is not None:
            self.fs.get(join_path(remote_artifact_path, "*"), to, recursive=True)
            return to
        else:
            return TmpArtifact(self, collection, artifact, version)



import json
import logging
import os

import wandb

WANDB_SECRET = "wandb-access-secret"
WANDB_KEY = "key_map"

logger = logging.getLogger(__name__)


def wandb_environment_setup(username: str) -> None:
    """
    Retrieve wandb login secrets secrets and export them as environment variables.
    This is a convenience function for our Flyte setup to be called before `wandb.init()`.

    If this function is called in local setup, i.e. outside of flyte task context, it will not do anything.

    Args:
        username (str): WandB user handle for look up of API key
    """
    # locally authentication should already be setup
    if wandb.api.api_key is None and "WANDB_API_KEY" not in os.environ.keys():
        # in a cloud context the secret containing the mapping usernames <> API-key should be accessible
        try:
            import flytekit

            secret_val = flytekit.current_context().secrets.get(WANDB_SECRET, WANDB_KEY)
            mapping = json.loads(secret_val)
            logger.info(f"Usernames found in the cloud: {list(mapping.keys())}")
            os.environ["WANDB_API_KEY"] = mapping[username]
        except (ModuleNotFoundError, ValueError):
            logger.warning(
                f"Could not retrieve secret: {WANDB_SECRET} {WANDB_KEY}\nNot changing setup - if wandb "
                f"authentication is an issue, please ensure the secret is made accessible. If you are executing "
                "locally, please set up the wandb environment variables manually."
            )



import logging
import operator
import os
import subprocess
import threading
from functools import reduce
from pathlib import Path
from typing import Dict, List, Optional

import docker

logger = logging.getLogger(__name__)


BUILD_TIMEOUT = 20 * 60  # 20 Mins for now


def early_termination(process, msg):
    """Terminate a process early, raise with the provided error message"""
    process.kill()
    raise Exception(msg)


def build_image(
    dockerfile: Path,
    workdir_path: Path,
    image_repo: str,
    image_tag: str,
    buildargs: Dict[str, str],
    secrets: Optional[List[str]] = None,
    push_image: bool = False,
    **kwargs: Dict[str, str],
) -> str:
    """Build docker image with local docker engine.

    Args:
        dockerfile (Path): path to the dockerfile
        workdir_path (Path): path to the working directory/docker context
        image_repo (str): fqn of the image repository. e.g eu.gcr.io/<proj_name>/<repo>
        image_tag (str): tag of the image. e.g <commit-hash>-dev
    Kwargs:
        buildargs (Dict[str,str]): Build args for docker build command
        secrets (List[str]): List of secret names to be used in the build
            Example, to mount google credentials from the host machine to the build context,
            you can pass `secrets=["id=gcp-cred,src=<path-to-creds-file>"]`
        push_image (bool, optional): flag to push the image to remote repo. Defaults to False.
        kwargs (Dict[str,str]): Additional arguments for docker build CLI call
    Returns:
        str: fqn:tag of the image

    """
    if image_repo[-1] == "/":
        raise ValueError("Image repo string should not end with '/'")
    docker_client = docker.from_env(timeout=BUILD_TIMEOUT)
    full_image_name = f"{image_repo}:{image_tag}"

    msg = f"Build image `{full_image_name}` with Dockerfile `{dockerfile}` and context `{workdir_path}`.\n"
    for key, val in buildargs.items():
        msg += f"Use ARG {key} `{val}`.\n"
    logger.info(msg)

    # Overwrite default values if they are not provided
    timeout = kwargs.pop("timeout", BUILD_TIMEOUT)

    # Use cli call to take advantage of BUILDKIT features
    os.environ["DOCKER_BUILDKIT"] = "1"

    docker_build_command = ["docker", "build", "-f", dockerfile, "--tag", full_image_name, "--progress=plain"]
    for key, val in kwargs.items():
        docker_build_command.extend([f"--{key}={val}"])
    for key, val in buildargs.items():
        docker_build_command.extend(["--build-arg", f"{key}={val}"])
    if kwargs.pop("pull", True):
        docker_build_command.append("--pull")
    if secrets:
        for secret in secrets:
            docker_build_command.append(f"--secret {secret}")

    docker_build_command.append(f"{workdir_path}")

    try:
        process = subprocess.Popen(
            " ".join(docker_build_command), shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, text=True
        )

        timer = threading.Timer(timeout, lambda: early_termination(process, f"Timeout reached after {timeout} sec"))
        timer.start()
        while True:
            line = process.stdout.readline()
            if not line:
                break
            logger.info("\t" + line.strip())
        timer.cancel()
    except subprocess.CalledProcessError as sub_err:
        logger.info(f"Docker command failed with return code: {sub_err.returncode} and msg\n{sub_err.stdout}")
        raise sub_err

    images = docker_client.api.images(name=image_repo)
    image_tags = [image.get("RepoTags", []) for image in images]
    if len(images) == 0 or full_image_name not in reduce(
        operator.concat, [tag for tag in image_tags if tag is not None]
    ):
        raise Exception(f"Build of image {full_image_name} not successful.")

    if push_image:
        logger.info(f"Pushing image `{full_image_name}`.")
        r = docker_client.images.push(full_image_name, stream=True, decode=True)
        for r_line in r:
            if "error" in r_line and r_line["error"]:
                raise docker.errors.APIError(r_line["error"])

    return full_image_name



from __future__ import annotations

import inspect
import queue
from abc import abstractmethod
from concurrent.futures import Executor, ThreadPoolExecutor
from copy import deepcopy
from functools import partial
from typing import Any, Callable, Generator, Iterable, Iterator, List, Optional, Type, Union

from scaffold.data.iterstream.iterators import batched_, filter_, flatten_, map_, shuffle_, take_, tqdm_

__all__ = ["Composable", "AsyncContent"]


class Composable(Iterable):
    """A mix-in class that provides stream manipulation functionalities."""

    def __init__(self, source: Optional[Union[Iterable, Callable]] = None):
        """Init"""
        self.source = source

    @abstractmethod
    def __iter__(self) -> Iterator:
        """Abstract iter"""
        pass

    def compose(self, constructor: Type[Composable], *args, **kw) -> Composable:
        """
        Apply the transformation expressed in the `__iter__` method of the `constructor` to items in the stream.
        If the provided constructor has an __init__ method, then the source argument should not be provided.
        """
        assert "source" not in kw
        if "__init__" in constructor.__dict__ and "source" in inspect.signature(constructor).parameters:
            raise ValueError(
                "If the provided constructor argument has an __init__ method, it should not "
                "have the source argument."
            )
        return constructor(*args, **kw).source_(self)

    def to(self, f: Callable, *args, **kw) -> _Iterable:
        """Pipe the iterable into another iterable which applies `f` callable on it"""
        assert "source" not in kw
        return _Iterable(self, f, *args, **kw)

    def source_(self, source: Union[Iterable, Callable]) -> Composable:
        """Set the source of the stream"""
        self.source = source
        return self

    def map(self, callback: Callable, **kw) -> _Iterable:
        """Applies the `callback` to each item in the stream. Specify key-word arguments for callback in **kw"""
        partial_callback = partial(callback, **kw)
        return self.to(map_, partial_callback)

    def filter(self, predicate: Callable) -> _Iterable:
        """Filters items by `predicate` callable"""
        return self.to(filter_, predicate)

    def async_map(
        self,
        callback: Callable,
        buffer: int = 100,
        max_workers: Optional[int] = None,
        executor: Optional[Executor] = None,
        **kw,
    ) -> _AsyncMap:
        """
        Applies the `callback` to the item in the self and returns the result.

        Args:
            callback (Callable): a callable to be applied to items in the stream
            buffer (int): the size of the buffer
            max_workers (int): number of workers in the
                :py:class:`ThreadPoolExecutor <concurrent.futures.ThreadPoolExecutor>`.
                `max_workers` is only used when `executor` is not provided, as the `executor`
                already includes the number of `max_workers`.
            executor (concurrent.futures.Executor, dask.distributed.Client): an optional executor to be used.
                By default a :py:class:`ThreadPoolExecutor <concurrent.futures.ThreadPoolExecutor>`
                is created, if no executor is provided. If you need a
                :py:class:`ProcessPoolExecutor <concurrent.futures.ProccessPoolExecutor>`,
                you can explicitly provide it here. It is also useful when chaining multiple
                `async_map`; you can pass the same `executor` to each `async_map` to share resources. If
                `dask.distributed.Client` is passed, tasks will be executed with the provided client (local or remote).

                **Note** if the executor is provided, it will not be closed in this function even after the iterator
                is exhausted.

                **Note** if executor is provided, the argument `max_workers` will be ignored. You should
                specify this in the executor that is being passed.
            **kw (dict): key-word arguments for callback

        Returns (_AsyncMap)
        """
        partial_callback = partial(callback, **kw)
        return _AsyncMap(
            source=self, callback=partial_callback, buffer=buffer, max_workers=max_workers, executor=executor
        )

    def flatten(self) -> _Iterable:
        """When items in the stream are themselves iterables, flatten turn them back to individual items again"""
        return self.to(flatten_)

    def batched(
        self, batchsize: int, collation_fn: Optional[Callable] = None, drop_last_if_not_full: bool = True
    ) -> _Iterable:
        """Batch items in the stream.

        Args:
            batchsize: number of items to be batched together
            collation_fn: Collation function to use.
            drop_last_if_not_full (bool): if the length of the last batch is less than the `batchsize`, drop it
        """
        return self.to(
            batched_,
            batchsize=batchsize,
            collation_fn=collation_fn,
            drop_last_if_not_full=drop_last_if_not_full,
        )

    def sliding(
        self,
        window_size: int,
        *,
        deepcopy: bool,
        stride: int = 1,
        drop_last_if_not_full: bool = True,
        min_window_size: int = 1,
        fill_nan_on_partial: bool = False,
    ) -> Composable:
        """
        Apply sliding window over the stream.

        Args:
            window_size (int): the length of the window
            deepcopy (bool): If True, each window will be returned as a deepcopy. If items are mutated in the
                subsequent steps of the pipeline, this should be set to True, otherwise it should be False.
                Note that deepcopy may incur a substantial cost, so set this parameter carefully.
            stride (int): the distance that the window moves at each step
            drop_last_if_not_full (bool): If True, it would only return windows of size `window_size` and drops the
                last items which have fewer items.
            min_window_size (int): The minimum length of the window for the last remaining elements.
                This argument is only relevant if `drop_last_if_not_full` is set to False, otherwise it's ignored.
            fill_nan_on_partial (bool): If `drop_last_if_not_full` is False, the length of the last few windows
                will be less than `window_size`. This argument fill the missing values with None if set to True.
                This argument take precedence over If `min_window_size`.

        """
        if not window_size > 1:
            raise ValueError("window_size must be > 1")
        if not window_size > min_window_size >= 1:
            raise ValueError("window_size must be greater than min_window_size, and min_window_size >= 1")
        if not window_size >= stride >= 1:
            raise ValueError("stride should be smaller or equal to window_size, and greater or equal to 1")

        return _SlidingIter(
            source=self,
            window_size=window_size,
            deepcopy=deepcopy,
            stride=stride,
            drop_last_if_not_full=drop_last_if_not_full,
            min_window_size=min_window_size,
            fill_nan_on_partial=fill_nan_on_partial,
        )

    def shuffle(self, size: Optional[int] = 1000, **kw) -> Composable:
        """Shuffles items in the buffer, defined by `size`, to simulate IID sample retrieval.

        Args:
            size (int, optional): Buffer size for shuffling. Defaults to 1000. Skip the shuffle step if `size < 2`.

        Acceptable keyword arguments:

        - initial (int, optional): Minimum number of elements in the buffer before yielding the first element.
          Must be less than or equal to `size`, otherwise will be set to `size`. Defaults to 100.

        - rng (random.Random, optional): Either `random` module or a :py:class:`random.Random` instance. If None,
          a `random.Random()` is used.

        - seed (Union[int, float, str, bytes, bytearray, None]): A data input that can be used for `random.seed()`.

        """

        if size is None:
            size = 1000

        if size < 2:
            return self
        return self.to(shuffle_, size, **kw)

    def take(self, n: Optional[int]) -> Composable:
        """Take n samples from iterable"""
        if n is None:
            return self
        return self.to(take_, n)

    def loop(self, n: Optional[int] = None) -> Composable:
        """Repeat the iterable n times.

        Args:
            n (int, Optional): number of times that the iterable is looped over. If None (the default), it loops forever

        Note: this method creates a deepcopy of the `source` attribute, i.e. all steps in the chain of Composables
        `before` the loop itself, which must be picklable.
        """
        return _LoopIterable(self, n)

    def zip_index(self, pad_length: int = None) -> Composable:
        """Zip the item in the stream with its index and yield Tuple[index, item]

        Args:
            pad_length: if provided, all indexes will be padded with zeros if they have less digits than pad_length,
                in which case all indexes are str rather than int.
        """
        return _ZipIndexIterable(self, pad_length=pad_length)

    def join(self) -> None:
        """A method to consume the stream"""
        for _ in self:
            pass

    def collect(self) -> List[Any]:
        """Collect and returns the result of the stream"""
        return list(self)

    def tqdm(self, **kw) -> _Iterable:
        """Add tqdm to iterator."""
        return self.to(tqdm_, **kw)


class _Iterable(Composable):
    """
    A class representing an iterable, which applies the callable `f` to the `source` items and returns the result in
    the :py:meth:`__iter__` method. This is used as the object being passed between steps of the stream.
    """

    def __init__(self, source: Iterable, f: Callable[..., Iterator], *args, **kw):
        """Initialize _Iterable.

        Args:
            source (Iterable): An iterable representing the source items.
            f (Callable): A callable to be applied to the iterator, which is built from `source`. Must return an
                iterator.
            *args: Arguments being passed to `f`.
            **kw: Kwargs passed to `f`.
        """
        super().__init__(source)
        assert callable(f)
        self.f = f
        self.args = args
        self.kw = kw

    def __iter__(self) -> Iterator:
        """Returns the iterator that is obtained by applying `self.f` to `self.source`."""
        assert self.source is not None, f"must set source before calling iter {self.f} {self.args} {self.kw}"
        assert callable(self.f), self.f
        return self.f(iter(self.source), *self.args, **self.kw)


class _SlidingIter(Composable):
    def __init__(
        self,
        source: Iterable,
        window_size: int,
        deepcopy: bool,
        stride: int = 1,
        drop_last_if_not_full: bool = True,
        min_window_size: int = 1,
        fill_nan_on_partial: bool = False,
    ):
        """Init"""
        super().__init__(source=source)
        self.window_size = window_size
        self.deepcopy = deepcopy
        self.stride = stride
        self.drop_last_if_not_full = drop_last_if_not_full
        self.min_window_size = min_window_size
        self.fill_nan_on_partial = fill_nan_on_partial

    def __iter__(self):
        it = iter(self.source)
        _win = []

        while len(_win) < self.window_size:
            try:
                _win.append(next(it))
            except StopIteration:
                if not self.drop_last_if_not_full:
                    yield from self._yield(self._fill_na(_win))
                return

        while True:
            yield from self._yield(_win)
            _win = self._step(_win, it)
            if (
                _win is None
                or (len(_win) < self.min_window_size and not self.fill_nan_on_partial)
                or all([i is None for i in _win])
            ):
                return

    def _step(self, win_: List, it_: Iterable) -> List | None:
        _new_items = []
        for _ in range(self.stride):
            try:
                _new_items.append(next(it_))
            except StopIteration:
                if not self.drop_last_if_not_full:
                    return self._fill_na(win_[self.stride :] + _new_items)
                else:
                    return
        return win_[self.stride :] + _new_items

    def _yield(self, _win: List) -> Generator[List[Any], None, None]:
        if self.deepcopy:
            yield deepcopy(_win)
        else:
            yield _win

    def _fill_na(self, _win: List) -> List | None:
        if all([i is None for i in _win]):
            return
        if self.fill_nan_on_partial and len(_win) < self.window_size:
            return _win + [None for _ in range(self.window_size - len(_win))]
        else:
            return _win


class _LoopIterable(Composable):
    def __init__(self, source: Iterable, n: Optional[int]):
        """Init"""
        super().__init__(source=source)
        self.n = n
        self.counter = 0

    def __iter__(self) -> Iterator:
        """Iterate over the iterable n times"""
        _started = False
        if self.n is None:
            current_ = iter(deepcopy(self.source))
            while True:
                try:
                    yield next(current_)
                    _started = True
                except StopIteration:
                    if not _started:
                        return
                    self.counter += 1
                    current_ = iter(deepcopy(self.source))
        else:
            for _ in range(self.n):
                yield from iter(deepcopy(self.source))
                self.counter += 1


class _ZipIndexIterable(Composable):
    def __init__(self, source: Iterable, pad_length: int = None) -> None:
        """Init"""
        super().__init__(source)
        self.idx = 0
        self.pad_length = pad_length

    def __iter__(self) -> Iterator:
        """Zip the index and the data"""
        for i in self.source:
            yield self._next_idx(), i

    def _next_idx(self) -> Union[int, str]:
        _idx = None
        if self.pad_length is not None:
            str_idx = str(self.idx)

            _idx = "0" * (self.pad_length - len(str_idx)) + str_idx
        else:
            _idx = self.idx
        self.idx += 1
        return _idx


class _AsyncMap(Composable):
    def __init__(
        self,
        source: Iterable,
        callback: Callable,
        buffer: int = 100,
        max_workers: Optional[int] = None,
        executor: Optional[Executor] = None,
    ):
        """A class that applies a `callback` asynchronously to the items in the `dataset`, using thread pool executor"""
        super().__init__(source)
        self.buffer = buffer
        self.callback = callback
        self.max_workers = max_workers
        self.executor = executor

        # Instantiate queue lazily in the __iter__ method
        # This is necessary to be compatible with the thread forking of PyTorch multiprocessing context
        # when using a multi-worker dataloader.
        self.queue = None

    def __iter__(self) -> Iterator:
        """An iterator"""

        self.queue = queue.Queue(self.buffer)
        it = iter(self.source)

        if self._executor_not_provided():
            with ThreadPoolExecutor(max_workers=self.max_workers) as exec_:
                yield from self._iter(it, exec_)
        elif isinstance(self.executor, Executor):
            yield from self._iter(it, self.executor)
        else:
            raise ValueError(f"Executor {self.executor} not recognized")

    def _executor_not_provided(self) -> bool:
        return self.executor is None

    def _iter(self, it: Iterator, executor: Executor) -> Iterator:
        sentinel = object()
        while True:
            # Fill queue
            while not self.queue.full():
                item = next(it, sentinel)
                if item is sentinel:
                    break
                self.queue.put(AsyncContent(item=item, func=self.callback, executor=executor))

            # stop iterating if all samples processed
            if self.queue.empty():
                break

            # yield sample
            yield self.queue.get().value()

    def _dask_iter(self, it: Iterator) -> Iterator:
        sentinel = object()
        while True:
            while not self.queue.full():
                item = next(it, sentinel)
                if item is sentinel:
                    break
                self.queue.put(self.executor.submit(self.callback, item))

            if self.queue.empty():
                break

            yield self.queue.get().result()


class AsyncContent:
    """Represents content that can be fetched asynchronously."""

    def __init__(self, item: str, func: Callable, executor: Executor) -> None:
        """Initialize AsyncContent.

        Args:
            item (str): Key corresponding to a single item, will be passed to `fetch_func`.
            func (Callable): Function that fetches a given key.
            executor (concurrent.futures.Executor): Executor to submit `func` with `item`.
        """
        self.stack = 1
        self.future = executor.submit(func, item)

    def value(self, timeout: int = None) -> Any:
        """Get the value asynchronously.

        Args:
            timeout (int, optional): Number of seconds to wait for the result. If None, then the future is waited
                indefinitely. Defaults to None.

        Returns:
            Any: Content.
        """
        return self.future.result(timeout)



"""
This module helps you managing hydra configs and running hydra sessions with configs.
Having an own runner for hydra session as a replacement for the @hydra.main annotation
is necessary since the original annotation performs parsing of argparse parameters and
running the sessions at the same time. To use e.g. click instead of argparse both aspects had to be disentangled.
"""

import logging
import os
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Tuple

from hydra._internal.config_search_path_impl import ConfigSearchPathImpl
from hydra._internal.hydra import Hydra as HydraClass
from hydra._internal.utils import detect_calling_file_or_module_from_task_function, run_and_report
from hydra.core.config_store import ConfigStore
from hydra.types import TaskFunction
from omegaconf import DictConfig

logger = logging.getLogger(__name__)


def resolve_rel_config_path(a_file: str, config_path: str) -> str:
    """Returns absolute path between relative file and folder

    Args:
        a_file: Reference file path e.g. __file__
        config_path: Relative path from e.g. __file__ to config

    Returns: str with absolute path to config
    """
    a_path = os.path.dirname(os.path.realpath(a_file))
    return str((Path(a_path) / config_path).resolve())


def resolve_and_split_path(config_path: str) -> Tuple[Path, str]:
    """Check if the file exists and split the path into the directory and base name without .yaml.

    Args:
        config_path: Path to the hydra config yaml.

    Returns:
        The absolute directory name and the file path without .yaml or .yml

    Raises:
        FileNotFoundError if file does not exist
    """
    config_path = Path(config_path).resolve()

    # check if the file exists, if not there is no error, but Hydra just uses the defaults
    # defined in the config store and ignores that you gave it a non-existing config path
    if not config_path.is_file():
        raise FileNotFoundError(f"Cannot find specified config file at {config_path}")

    return config_path.parent, config_path.name.replace(".yaml", "").replace(".yml", "")


def run_with_hydra(task_function: TaskFunction, config_dir: str, config_name: str, overrides: List[str] = None) -> Any:
    """Run a function in a hydra session using the specified config.

    Args:
        task_function: Method that should be executed.
        config_path: Absolute path to config.
        overrides: List of config parameter overrides with
                   syntax https://hydra.cc/docs/next/advanced/override_grammar/basic/"

    Returns:
        Return value from task_function
    """
    from hydra.core.global_hydra import GlobalHydra

    if overrides is None:
        overrides = []

    # create a hydra session
    hydra_env = _get_hydra_env(task_function, config_dir)

    # adapted from https://github.com/facebookresearch/hydra/blob/bf1f92c49ede2db7b3e2e70a06c68c3a6b0bfe41/hydra/_internal/utils.py#L274 # noqa: E501
    ret = None
    try:
        # run task_function using hydra
        ret = run_and_report(
            lambda: hydra_env.run(config_name=config_name, task_function=task_function, overrides=overrides)
        )
    finally:
        # cleanup hydra session
        GlobalHydra.instance().clear()

    return ret.return_value


def _get_hydra_env(task_function: TaskFunction, config_dir: str) -> HydraClass:
    """Discovers provided, plugin, and global configs and bootstraps a hydra session with them.

    Args:
        task_function: File path to the yaml config file. This file path does not have to include the ".yaml" extension.
        config_dir: Absolute path to the config directory

    Returns:
        Return hydra instance
    """
    # adapted from https://github.com/facebookresearch/hydra/blob/bf1f92c49ede2db7b3e2e70a06c68c3a6b0bfe41/hydra/_internal/utils.py#L274 # noqa: E501

    # lazy imports to not slow down components that do not use hydra
    from hydra.core.plugins import Plugins
    from hydra.plugins.search_path_plugin import SearchPathPlugin

    # retrieve attributes of task_function to later name the hydra run after it
    (
        calling_file,
        calling_module,
        task_name,
    ) = detect_calling_file_or_module_from_task_function(task_function)

    # bootstrap config discovery
    search_path = ConfigSearchPathImpl()

    # add specified config based on an absolute path
    search_path.append("command-line", f"file://{config_dir}")

    # add global hydra config
    search_path.append("hydra", "pkg://hydra.conf")

    # add configs found by installed SearchPathPlugins
    search_path_plugins = Plugins.instance().discover(SearchPathPlugin)
    for spp in search_path_plugins:
        plugin = spp()
        assert isinstance(plugin, SearchPathPlugin)
        plugin.manipulate_search_path(search_path)

    # add installed schema discoveries
    search_path.append("schema", "structured://")

    # use run_and_report to execute a hydra run which returns a hydra run environment
    hydra_env = run_and_report(
        lambda: HydraClass.create_main_hydra2(task_name=task_name, config_search_path=search_path)
    )

    # INFO: Do not forget to cleanup your session with GlobalHydra.instance().clear()
    return hydra_env


def init_configstore(config_name: str, config_schema: "dataclass") -> ConfigStore:
    """Method to initialise the configstore. To be called if the entrypoint gets executed as the config file should then
    adhere to this schema.

    Args:
        config_name: Note that this name MUST be equal to the config file (without .yaml) name, otherwise Hydra doesn't
            merge the config yaml with this config store.
        config_schema: Schema dataclass definition for the config.

    Returns: None

    Raises:
        ValueError: If `config_name` contains the file extension.
    """

    if Path(config_name).suffix != "":
        raise ValueError(f"File extension must be omitted in 'config_name'. Given: {config_name}.")
    cs = ConfigStore.instance()
    dataclass_name = config_schema.__name__
    # we set the name of the dataclass as name of the top level schema
    # that means if we look at a bundle yaml, we know which dataclass defines the schema
    cs.store(name=dataclass_name, node=config_schema)
    return cs


def get_cfg_from_config_fp(
    config_fp: str,
    config_class: Optional["dataclass"] = None,  # noqa: F821
    overrides: List[str] = None,
    add_schema_funcs: Optional[List[Callable]] = None,
) -> DictConfig:
    """
    Helper function that returns a hydra config object.
    Can be used for tests or debugging and the show config command.

    Before composing the config provided in `config_fp`, all functions in `add_schema_funcs` will be called.

    IMPORTANT:
        This does not work when Hydra is already initialized

    Args:
        config_fp (str): absolute filepath to the component config (bundle).
        config_class (Optional["dataclass"]): Dataclass schema for the config.
        overrides (List[str] = None): Values in the config that should be overridden.
        add_schema_funcs (Optional[List[Callable]] = None): Functions that should be called after the
            config store is initialised. These functions can be used to add schemas to the config store
            so they can be checked against the given config file. Since these schemas can vary between
            components, every component implements these separately.

    Returns:
        The dataclass representing the config object

    Raises:
        FileNotFoundError: If specified config file does not exist.
    """
    # This function is not needed for most entrypoints, so just import the stuff we need for it lazily
    from hydra import compose, initialize_config_dir

    overrides = overrides if overrides is not None else []
    add_schema_funcs = add_schema_funcs if add_schema_funcs is not None else []

    config_dir, config_name = resolve_and_split_path(config_fp)

    with initialize_config_dir(config_dir=str(config_dir.absolute())):
        # Load the schema
        if config_class is not None:
            cs = init_configstore(config_name, config_class)
        else:
            cs = ConfigStore.instance()
        for add_schema in add_schema_funcs:
            add_schema(cs)
        cfg = compose(config_name, overrides=overrides)

    return cfg


def run_component_with_config_fp(
    run_component_func: Callable,
    config_fp: str,
    config_bundle_class: Optional["dataclass"] = None,  # noqa: F821
    overrides: List[str] = None,
    add_schema_funcs: Optional[List[Callable]] = None,
) -> Dict:
    """
    Generic entry point for an entrypoint component based on a config file path.
    This function should standardize how to instantiate the config store, the hydra environment
    and finally run the given function with the config in that environment.

    Before running the component, all functions in `add_schema_funcs` will be called which should
    add all needed schemas to the hydra config store.

    IMPORTANT:
        This call will try to instantiate a new Hydra environment, so calling this function inside an
        environment where a global Hydra instance already exists will raise an Error.
        If you want to call an entrypoint component inside an existing Hydra state (i.e. in another project
        that also used hydra), please try to load the needed configs for that component manually.

    Args:
        run_component_func (Callable): Component function that takes a config instance.
        config_fp (str): absolute filepath to the component config (bundle).
        config_bundle_class (Optional["dataclass"]): Dataclass schema for the config.
        overrides (List[str] = None): Values in the config that should be overridden.
        add_schema_funcs (Optional[List[Callable]] = None): Functions that should be called after the
            config store is initialised. These functions can be used to add schemas to the config store
            so they can be checked against the given config file. Since these schemas can vary between
            components, every component implements these separately.

    Returns:
        Return value of the run_component_func
    """
    overrides = overrides if overrides is not None else []
    add_schema_funcs = add_schema_funcs if add_schema_funcs is not None else []

    config_dir, config_name = resolve_and_split_path(config_fp)
    if config_bundle_class is not None:
        cs = init_configstore(config_name, config_bundle_class)
    else:
        cs = ConfigStore.instance()
    for add_schema in add_schema_funcs:
        add_schema(cs)

    return run_with_hydra(run_component_func, str(config_dir), config_name, overrides)



import typing as t
from dataclasses import field

from omegaconf import MISSING

from scaffold.hydra.config_helpers import structured_config

GROUP = "scaffold/artifact_manager"


@structured_config(group=GROUP)
class ArtifactManagerConf:
    _target_: str = MISSING
    collection: str = "default"


@structured_config(group=GROUP)
class WandbArtifactManagerConf(ArtifactManagerConf):
    _target_: str = "scaffold.data.artifact_manager.wandb.WandbArtifactManager"
    entity: str = "mxm"
    project: str = MISSING


@structured_config(group=GROUP)
class FileSystemArtifactManagerConf(ArtifactManagerConf):
    _target_: str = "scaffold.data.artifact_manager.filesystem.FileSystemArtifactManager"
    url: str = MISSING
    fs_kwargs: t.Dict = field(default_factory=dict)



import logging
import random
from typing import MutableMapping, Optional, Union

import random_name
from fsspec.spec import AbstractFileSystem

from scaffold.data.artifact_manager.base import ArtifactManager

FILESYSTEM = Union[MutableMapping, AbstractFileSystem]

logger = logging.getLogger(__name__)


def validate_or_generate_target_afid(
    artifact_manager: ArtifactManager, afid: Optional[str] = None, prefix: Optional[str] = None
) -> str:
    """Validates a given target afid or generates a new one.

    Takes the given afid and checks its validity, raises a ValueError if not. If the given afid is None, a new afid
    is returned with the optional prefix. Can be used when validating any afid parameters that are meant for
    targets. Target afids refer to the ids for artifacts that will be generated by your entrypoint and therefore
    should not exist yet, in contrast to "input afids" which are ids for artifacts an entrypoint needs as input to
    do something, and therefore _should_ exist in the artifact store.

    Args:
        afid: An artifact id that is supposed to be not used yet on the server, default None.
        prefix: Only used when afid is None. Will be the prefix of the new afid.

    Raises:
        ValueError if an afid was given, but it already exists in the artifact store.

    Returns:
        Either the given afid if its valid, or a new one.
    """
    if afid is None:
        afid = get_new_afid(artifact_manager=artifact_manager, prefix=prefix)
        logger.info(f"Artifact id was not set in config. Got new unique id: {afid}")
    else:
        if artifact_manager.exists(afid):
            raise ValueError(
                f"The given artifact id ({afid}) to store the result would not be unique in this artifact store"
                f". Please specify an artifact id that is unique across ALL runs of the artifact store,"
                f" or leave it empty to get a new id assigned."
            )
    return afid


def get_new_afid(artifact_manager: ArtifactManager, prefix: str = None) -> str:
    """Get a new artifact id.

    The identifiers are triplets of random words, to make them more distinguishable for the human eye, plus the
    timestamp to ensure uniqueness. Optionally, a prefix can be added to the id.

    Args:
        prefix: String to be added to the uid, empty by default

    Returns:
        String in form of prefix__new_id__timestamp
    """
    import time

    random.seed()  # Overwrite the seed set by torch_lightning.utilities.seed.seed_everything during training.
    timestamp = time.strftime("%Y-%m-%d_%H-%M-%S")
    new_id = (prefix + "__" if prefix is not None else "") + random_name.generate_name() + "__" + timestamp

    return new_id


def validate_input_afid(artifact_manager: ArtifactManager, afid: str) -> None:
    """Validates input afid, i.e. an afid that is meant to be existing in the artifact store, as it is needed as
    input to a entrypoint.

    Args:
            afid: An artifact id that is supposed to be existing in the artifact store.

    Raises:
            ValueError if an afid was given, but it does not exist in the artifact store.
    """
    if not artifact_manager.exists(afid):
        raise ValueError(f"The given artifact id ({afid}) does not exist in the artifact store.")



import os
from pathlib import Path

from docker.utils.build import PatternMatcher
from flytekit.tools.ignore import Ignore


class FlyteIgnore(Ignore):
    def __init__(self, root: Path, ignore_path: Path):
        """Altered version of Flytes DockerIgnore allowing custom path. Relies on Dockers PatternMatcher."""
        super().__init__(root)
        self.pm = self._parse(ignore_path)

    def _parse(self, path: Path) -> PatternMatcher:
        """Parse a file in the style of dockerignore and return a PatternMatcher"""
        patterns = []
        dockerignore = os.path.join(self.root, path)
        if os.path.isfile(dockerignore):
            with open(dockerignore, "r") as f:
                patterns = [line.strip() for line in f.readlines() if line and not line.startswith("#")]
        return PatternMatcher(patterns)

    def _is_ignored(self, path: str) -> bool:
        """Check if a path is ignored by the ignore file"""
        return self.pm.matches(path) and path != ""



import importlib
import logging
from contextlib import contextmanager
from typing import Callable, Dict, Generator, List, Set, Tuple, Union

from flytekit.configuration import (
    Config,
    FastSerializationSettings,
    Image,
    ImageConfig,
    PlatformConfig,
    SerializationSettings,
)
from flytekit.core.base_task import PythonTask
from flytekit.core.condition import BranchNode
from flytekit.core.node import Node
from flytekit.core.workflow import WorkflowBase
from flytekit.remote.remote import FlyteRemote

logger = logging.getLogger(__name__)


def mxm_register(nodes: List[Union[PythonTask, WorkflowBase]]) -> Callable:
    """
    Wrapper for both functions and workflows to declare additional flyte components for registration.
    This should go outside the flyte decorators (task, dynamic, workflow etc.)

    Args:
        nodes (List[Union[PythonTask, WorkflowBase]]): A list of tasks and workflows to be registered

    Returns:
        Union[PythonTask, WorkflowBase]: The original workflow
    """

    def decorator(func: Callable) -> Callable:
        assert isinstance(func, PythonTask) or isinstance(
            func, WorkflowBase
        ), f"Can only annotate Flyte entities not {type(func)}"
        func.mxm_nodes = nodes
        return func

    return decorator


def identify_main_workflow(module_name: str) -> Tuple[WorkflowBase, List[Union[PythonTask, WorkflowBase]]]:
    """
    Identify the main Flyte workflow and its subcomponents in a python module.

    Top-level workflows are all workflows defined in the same module as hydra.main.
    The main workflow is the only top-level workflow which is not used in any other top-level workflow - it is assumed,
    that exactly one such workflow exists.
    The required flyte components that also need to be registered are identified by recursively descending through
    the workflow's node tree.

    Note: The main workflow is not contained in the returned list itself.

    Args:
        module_name (str): the python module

    Returns:
        Tuple[WorkflowBase, List[Union[PythonTask, WorkflowBase]]]:
            The main workflow and a list of contained flyte entities to be registered

    Raises:
        ValueError: If the module does not contain exactly one workflow.
    """
    module = importlib.import_module(module_name)
    workflows_to_nodes: Dict[WorkflowBase, Set[Union[PythonTask, WorkflowBase]]] = dict(
        (getattr(module, attr), []) for attr in dir(module) if isinstance(getattr(module, attr), WorkflowBase)
    )

    sub_workflows = set()
    for wf in workflows_to_nodes.keys():
        if wf not in sub_workflows:
            wf_children = [node.flyte_entity for node in wf.nodes]
            if hasattr(wf, "mxm_nodes"):
                wf_children += wf.mxm_nodes
            entities = extract_flyte_entities(set(wf_children), set())
            sub_workflows.update([entity for entity in entities if isinstance(entity, WorkflowBase)])
            workflows_to_nodes[wf] = entities

    main_workflows = list(set(workflows_to_nodes.keys()).difference(sub_workflows))
    assert len(main_workflows) == 1, (
        "Can only register one main workflow. Please be more conservative with workflow "
        "imports and ensure that no unused sub-workflows are defined or imported."
    )
    return main_workflows[0], list(workflows_to_nodes[main_workflows[0]])


def extract_flyte_entities(
    new_entities: Set[Union[PythonTask, WorkflowBase]], entities: Set[Union[PythonTask, WorkflowBase]]
) -> Set[Union[PythonTask, WorkflowBase]]:
    """
    Recursively extract sub-workflows and other flyte entities from a list of nodes or flyte entities.

    Args:
        new_entities (List[Union[PythonTask, WorkflowBase]]): flyte entities used to recursively extract sub-entities
        entities (Set[Union[PythonTask, WorkflowBase]]): already discovered entities not to recurse on

    Returns:
        Set[Union[PythonTask, WorkflowBase]]:
            A set of all children nodes to be registered (including subworkflows).
    """
    # prepare next descend through recursion
    children: Set[Union[Node, PythonTask, WorkflowBase]] = set()
    for entity in new_entities:
        if isinstance(entity, PythonTask):
            # standard tasks, map-tasks and dynamic workflows - add to entities
            entities.add(entity)
        elif isinstance(entity, WorkflowBase):
            # standard workflows - add entity to entities list and add child nodes to next recursion
            children.update(set([node.flyte_entity for node in entity.nodes]))
            entities.add(entity)
        elif isinstance(entity, BranchNode):  # Flyte conditional statements
            clauses = [entity._ifelse_block._case._then_node, entity._ifelse_block._else_node]
            if entity._ifelse_block.other is not None:
                clauses.extend([elif_node._then_node for elif_node in entity._ifelse_block.other])
            children.update(set([clause.flyte_entity for clause in clauses]))
        else:
            raise ValueError(
                f"Encountered Flyte entity of type {type(entity)} during registration. This type is not known and will "
                f"not be registered."
            )

        if hasattr(entity, "mxm_nodes"):
            children.update(entity.mxm_nodes)

    # filter already discovered entities to avoid infinite regress
    children = children.difference(entities)

    if len(children) == 0:
        return entities
    else:
        return extract_flyte_entities(children, entities)


@contextmanager
def temp_flyte_remote(
    project: str,
    domain: str,
    endpoint: str,
) -> Generator[Tuple[FlyteRemote, SerializationSettings], None, None]:
    """
    Create a flyte remote object with a temporary proxy into the kubernetes cluster.

    Args:
        project (str): the flyte project
        domain (str): one of `development`, `staging` or `production`
        service (str): kubernetes service name to port-forward.
        port (int): port of the service in the cluster to port-forward
        namespace (str): namespace of the service in the cluster that is being port-forwarded
        local_port (int): the desired local port to forward to, will use a random port if the desired port is in use
        venv_root (str): python virtual environment root for the flyte image, flyte default is used when not specified.

    Yields:
        Generator[FlyteRemote, None, None]:
            FlyteRemote: flyte remote object to make register calls.
    """
    remote = FlyteRemote(
        default_project=project,
        default_domain=domain,
        config=Config(platform=PlatformConfig(endpoint=endpoint, insecure=True)),
    )

    yield remote


def get_serialization_settings(
    default_image: str,
    extra_images: Dict[str, str],
    fast_serialization_settings: FastSerializationSettings,
    project: str,
    domain: str,
) -> SerializationSettings:
    """
    Create flyte serialization settings for registering tasks and workflows.

    Args:
        default_image (str): the tag of the image used by default by the flyte tasks
        extra_images (Dict[str, str]): a dictionary mapping the names of extra images to be used to the respective tags.
            The format is e.g. `{"spark": "eu.gcr.io/project_name/spark_image:tag"}`.
            In the flyte task decorator, the image is specified e.g. with
            `container_image="{{.images.spark.fqn}}:{{.images.default.version}}"`.
        fast_serialization_settings (FastSerializationSettings): Details of image injections in fast serialisation mode
        project (str): the flyte project
        domain (str): domain, normally one of `development`, `staging` or `production
    Returns:
        Flyte SerializationSettings for registering flyte entities.
    """
    extra_image_configs = []

    for label, image in extra_images.items():
        fqn, tag = image.rsplit(":", 1)
        extra_image_configs.append(Image(name=label, fqn=fqn, tag=tag))

    default_image_fqn, default_image_tag = default_image.rsplit(":", 1)
    default_serialization_settings = SerializationSettings(
        image_config=ImageConfig(
            default_image=Image(name="default", fqn=default_image_fqn, tag=default_image_tag),
            images=extra_image_configs,
        ),
        fast_serialization_settings=fast_serialization_settings,
        project=project,
        domain=domain,
    )

    return default_serialization_settings



import os
from contextlib import contextmanager
from typing import Iterator, Optional

import hydra
from hydra.core.global_hydra import GlobalHydra

from scaffold.hydra.constants import VERSION_BASE


@contextmanager
def initialize(
    config_dir: Optional[str] = None, job_name: Optional[str] = "app", exists_ok: Optional[bool] = True
) -> Iterator[GlobalHydra]:
    """Context manager, that either uses an existing hydra instance, or initializes one temporarily.

    Args:
        config_dir (Optional[str]): Absolute directory that should be added to the search path.
        job_name (Optional[str]): hydra.job.name which should be set.
        exists_ok (Optional[bool]): Do not fail if hydra is already initialized and yield GlobalHydra.instance().
    """
    if config_dir is not None and not os.path.isabs(config_dir):
        raise ValueError("scaffold.hydra.initialize() requires an absolute config_dir!")
    if GlobalHydra.instance().is_initialized():
        if exists_ok:
            yield GlobalHydra.instance()
        else:
            raise ValueError(
                "GlobalHydra is already initialized, call GlobalHydra.instance().clear() if you want to re-initialize, "
                "or set scaffold.hydra.initialize(exist_ok=True) to return the current instance."
            )
    else:
        if config_dir is not None:
            with hydra.initialize_config_dir(config_dir=config_dir, job_name=job_name, version_base=VERSION_BASE):
                yield GlobalHydra.instance()
        else:
            with hydra.initialize(config_path=None, job_name=job_name, version_base=VERSION_BASE):
                yield GlobalHydra.instance()



from __future__ import annotations

from abc import abstractmethod
from copy import deepcopy
from dataclasses import dataclass
from functools import wraps
from pathlib import Path
from types import MethodType
from typing import Callable, Dict, Optional, Type

from hydra.core.config_store import ConfigStore

CONFIG_STORE = ConfigStore.instance()


class StructuredConfig:
    @classmethod
    @abstractmethod
    def get_store_args(cls: Type[StructuredConfig]) -> Dict[str, str]:
        """Returns the arguments used to store this dataclass in the config store."""
        ...

    @classmethod
    @abstractmethod
    def get_store_path(cls: Type[StructuredConfig]) -> str:
        """Returns the path to the node of the StructuredConfig inside the config store."""
        ...


def _add_structured_config_base(cls: "dataclass", store_args: dict) -> type:
    """Dynamically add StructuredConfig as a base to dataclass.

    This is done, because the scaffold.hydra.compose() method needs to know the config store location of the dataclass.
    Additionally this only requires the user to use @structured_config without inheriting from StructuredConfig.

    You are then able to compose a structured config after importing in python, without knowing its config store
    position.
    >>> from scaffold.hydra import compose
    >>> config = compose(MyStructuredConfig, overrides=[...])

    Args:
        store_args (Dict): Arguments for the config store location.

    Returns:
        New class with base StructuredConfig.
    """
    _store_args = deepcopy(store_args)

    def get_store_args(cls: StructuredConfig) -> Dict[str, str]:
        """Returns the arguments used to store this dataclass in the config store."""
        return deepcopy(_store_args)

    def get_store_path(cls: StructuredConfig) -> str:
        """Returns the path to the node of the StructuredConfig inside the config store."""
        _args = cls.get_store_args()
        if _args["group"] is None:
            return _args["name"]
        return str(Path(_args["group"], _args["name"]))

    cls.get_store_args = MethodType(get_store_args, cls)
    cls.get_store_path = MethodType(get_store_path, cls)

    # Add StructuredConfig as base. Inheritance order from right to left.
    # Also making sure that the resulting class is still attached to the original module.
    cls = type(cls.__name__, (cls, StructuredConfig), {"__module__": cls.__module__})
    return cls


def structured_config(
    _cls: Optional[Type] = None,
    name: Optional[str] = None,
    group: Optional[str] = None,
    package: Optional[str] = None,
    provider: Optional[str] = None,
    frozen: bool = False,
) -> Callable:
    """Decorator which extends the @dataclass in the following ways:
    1. Adds the decorated class to the config store with reasonable defaults for
       name (class name) and provider (python package)
    2. Adds the base type StructuredConfig to the class
    3. Adds a class method get_store_args() which can be used to get the store location from the class.

    NOTE: The decorated class is only registered, if the module containing it is imported!

    Example:
        >>> @structured_config
            class MyRootConfig()
                ...
        >>> @structured_config(group="my/group")
            class MyGroupConfig()
                ...

    Args:
        name (Optional[str]): Name of the config node. By default the class.__name__.
        group (Optional[str]): Config group, subgroup separator is '/'. If None, will be placed in the root.
        package (Optional[str]): Config package, which specifies where the config is applied.
            By default, this is derived from the group ("my/group" -> my.group)
        provider (Optional[str]): The name of the module/app providing this config.
            Helps debugging. By default this is set to the python module of the structured_config.
        frozen (Optional[bool]): Whether the dataclass should be read only.

    Returns:
        Class decorator.
    """

    @wraps(_cls, updated=())  # See wraps use for classes https://stackoverflow.com/a/65470430
    def struct_config_decorator(cls: type) -> type:
        dataclass_decorator = dataclass(frozen=frozen)
        cls = dataclass_decorator(cls)

        # When using "/group" hydra adds an empty key as a first level ["", "group"].
        # To keep things consistent, we always want to use the first group as the first key.
        _group = group.strip("/") if group is not None else group
        store_args = dict(
            name=name or cls.__name__,
            group=_group,
            package=package,
            provider=provider or cls.__module__,
        )

        cls = _add_structured_config_base(cls, store_args)

        # Note: node class for hydra store needs to be set after class extension with _add_structured_config_base
        store_args["node"] = cls

        CONFIG_STORE.store(**store_args)

        return cls

    # Decorator either used as @structured_config or @structured_config()
    return struct_config_decorator(_cls) if _cls else struct_config_decorator



from collections.abc import MutableMapping
from pathlib import Path
from typing import Any, Dict, Optional, Union

from fsspec.spec import AbstractFileSystem

URL = Union[str, Path]
FILESYSTEM = Union[MutableMapping, AbstractFileSystem]
SampleType = Dict[str, Any]
SeedType = Optional[Union[int, float, str, bytes, bytearray]]



import typing as t
from dataclasses import field

from omegaconf import MISSING

from scaffold.hydra.config_helpers import structured_config


@structured_config(group="ctx_manager")
class ContextConfig:
    _target_: str = MISSING


@structured_config(group="ctx_manager")
class WandBContextConfig(ContextConfig):
    _target_: str = "scaffold.ctx_manager.WandBContext"
    base_url: str = MISSING
    project: str = MISSING
    entity: str = "mxm"
    group: t.Optional[str] = None
    job_type: t.Optional[str] = None
    tags: t.Optional[t.List[str]] = None
    name: t.Optional[str] = None
    notes: t.Optional[str] = None
    run_id: t.Optional[str] = None
    user: t.Optional[str] = None
    resume: bool = False


@structured_config(group="ctx_manager")
class LoggingContextConfig(ContextConfig):
    defaults: t.List[t.Any] = field(
        default_factory=lambda: [
            {"scaffold/entrypoint/logging@log_config": "default"},
        ]
    )

    _target_: str = "scaffold.ctx_manager.LoggingContext"
    log_config: t.Dict[str, t.Any] = MISSING  # Same options as hydra.job_logging
    verbose: t.Any = False  # Same as hydra.verbose, but applied to our logging setup, instead of hydra.job_logging



from dataclasses import field
from enum import Enum, unique
from typing import Dict, List, Optional

from flytekit.models.core.execution import WorkflowExecutionPhase

from scaffold.hydra.config_helpers import structured_config

GROUP = "hydra/launcher"


@unique
class FlyteDomainEnum(str, Enum):
    development = "development"
    staging = "staging"
    production = "production"


@unique
class FlyteNotificationEnum(str, Enum):
    email = "email"
    slack = "slack"


@unique
class FlyteWorkglowExecutionPhaseEnum(str, Enum):
    UNDEFINED = WorkflowExecutionPhase.UNDEFINED
    QUEUED = WorkflowExecutionPhase.QUEUED
    RUNNING = WorkflowExecutionPhase.RUNNING
    SUCCEEDING = WorkflowExecutionPhase.SUCCEEDING
    SUCCEEDED = WorkflowExecutionPhase.SUCCEEDED
    FAILING = WorkflowExecutionPhase.FAILING
    FAILED = WorkflowExecutionPhase.FAILED
    ABORTED = WorkflowExecutionPhase.ABORTED
    TIMED_OUT = WorkflowExecutionPhase.TIMED_OUT
    ABORTING = WorkflowExecutionPhase.ABORTING


@unique
class ExecutionEnvironmentEnum(str, Enum):
    local = "local"
    remote = "remote"


@structured_config(group="scaffold/flyte_launcher")
class FlyteDockerImageConfig:
    # Repo and version tag injected as ARG BASE_IMAGE and ARG BASE_VERSION into docker build step
    # If image is not being build, this is the image that is being used
    base_image: str = "${hydra.launcher.workflow.default_image.base_image}"
    base_image_version: Optional[str] = "${hydra.launcher.workflow.default_image.base_image_version}"
    # target repo and tag the built container is pushed to
    target_image: str = "${hydra.launcher.workflow.default_image.target_image}"
    target_image_version: Optional[str] = "${hydra.launcher.workflow.version}"
    # Docker context is always relative to the project root, which is identified by the presence of the pyproject.toml
    # and other files like setup.py, requirements.txt, etc. (source: https://github.com/chendaniely/pyprojroot)
    dockerfile_path: str = "infrastructure/docker/Dockerfile"
    docker_context: str = "."
    # keyword args for build call: https://docker-py.readthedocs.io/en/stable/api.html#module-docker.api.build
    buildargs: Dict[str, str] = field(default_factory=dict)
    # mount secrets to the build context
    secrets: Optional[List[str]] = None
    docker_kwargs: Dict[str, str] = field(default_factory=dict)
    # Internal image name, that tasks refer to in flyte.
    flyte_image_name: str = "default"


@structured_config(group="scaffold/flyte_launcher")
class FlyteWorkflowConfig:
    default_image: FlyteDockerImageConfig = field(
        default_factory=lambda: FlyteDockerImageConfig(
            base_image="${hydra.launcher.workflow.default_image.target_image}",
            base_image_version="latest",
        )
    )  # Image used for all tasks by default.
    extra_images: List[FlyteDockerImageConfig] = field(default_factory=list)
    ignore: str = ".flyteignore"
    version: Optional[str] = None  # If not set, will be automatically created based on git hash or random string.
    project: str = "default"  # Flyte project that the workflow will be registered in.
    domain: FlyteDomainEnum = FlyteDomainEnum.development  # Flyte domain inside the flyte project.
    cron_schedule: Optional[str] = None  # Cron schedule for the execution of the workflow.


@structured_config(group="scaffold/flyte_launcher")
class FlyteNotificationConfig:
    type: FlyteNotificationEnum = FlyteNotificationEnum.email
    phases: List[FlyteWorkglowExecutionPhaseEnum] = field(default_factory=list)
    recipients: List[str] = field(default_factory=list)


@structured_config(group=GROUP, name="flyte")
class LauncherConfig:
    _target_: str = "hydra_plugins.flyte_launcher_plugin._flyte_launcher.FlyteLauncher"
    execution_environment: ExecutionEnvironmentEnum = ExecutionEnvironmentEnum.remote
    endpoint: str = "localhost:30081"
    build_images: bool = True
    fast_serialization: bool = False
    run: bool = True
    workflow: FlyteWorkflowConfig = field(default_factory=FlyteWorkflowConfig)
    notifications: List[FlyteNotificationConfig] = field(default_factory=list)



"""
At a low level, iterstream is defined through a class called :py:class:`Composable`.
All chaining methods are defined within this class, and can be applied at its high level twin class called
:py:class:`IterableSource`.
"""

from scaffold.data.iterstream.base import Composable
from scaffold.data.iterstream.source import FilePathGenerator, IterableSamplerSource, IterableSource

__all__ = [
    "Composable",
    "IterableSource",
    "IterableSamplerSource",
    "FilePathGenerator",
]



import random
from concurrent.futures import ThreadPoolExecutor
from typing import Callable, Iterable, Iterator, List, Optional, Union

from fsspec.core import url_to_fs

from scaffold.data.fs import get_protocol
from scaffold.data.iterstream.base import AsyncContent, Composable
from scaffold.data.iterstream.iterators import get_random_range

__all__ = ["IterableSource", "IterableSamplerSource"]


class IterableSource(Composable):
    """A class that turns an iterable to a source of a stream and provides stream manipulation functionalities on top,
    for instance:
    - map
    - map_async
    - filter
    - batched
    - shuffle
    - and more

    For the detailed description of each, please refer to the corresponding docstring in :py:class:`Composable`.
    """

    def __init__(self, source: Optional[Union[Iterable, Callable]] = ()):
        """Initialize IterableSource.

        Args:
            source (Union[Iterable, Callable], Optional): An Iterable that the IterableSource is
                built based on, or a callable that generates items when called.
        """
        super().__init__(source=source)

    def __iter__(self) -> Iterator:
        """Iterates over the items in the iterable"""
        if isinstance(self.source, Callable):
            yield from self.source()
        else:
            yield from self.source


class IterableSamplerSource(Composable):
    """A class that samples from iterables into an iterstream."""

    def __init__(
        self,
        iterables: List[Iterable],
        probs: Optional[List[float]] = None,
        rng: Optional[random.Random] = None,
        seed: Optional[int] = None,
    ):
        """Initialize IterableSamplerSource.

        Args:
            iterables (List[Iterable]): List of iterables to sample from.
            probs (Optional[List[float]], optional): [description]. Defaults to None.
            rng (random.Random, optional): Random number generator to use.
            seed (Optional[int]): An int or other acceptable types that works for random.seed(). Will be used to seed
                `rng`. If None, a unique identifier will be used to seed.
        """
        super().__init__(source=())
        self.rng = get_random_range(rng, seed)
        self.iterables = iterables
        if probs is not None:
            assert len(probs) == len(self.iterables), "number of iterables and probs must be equal"
            assert sum(probs) == 1, "sum of probs must add up to 1"
            assert all(p > 0 for p in probs), "probability for each iterable must be positive"
        self.probs = probs

    def __iter__(self) -> Iterator:
        """Samples items from the iterables, returns all samples until all iterables are exhausted."""
        iterators = [iter(it) for it in self.iterables]
        while True:
            idx = self.rng.choices(range(len(iterators)), weights=self.probs)[0]
            try:
                yield next(iterators[idx])
            except StopIteration:
                iterators = iterators[:idx] + iterators[idx + 1 :]
                if self.probs is not None:
                    probs = self.probs[:idx] + self.probs[idx + 1 :]
                    s = sum(probs)
                    self.probs = [p / s for p in probs]
            if len(iterators) == 0:
                break


class FilePathGenerator(Composable):
    """
    A specialized version of `Composable` that accepts a url without instantiating a filesystem instance in the init.
    It simply generates directories under the given `url` by instantiating a fsspec filesystem and yielding the result
    of fs.ls(url).
    """

    def __init__(
        self,
        url: str,
        nested: bool = False,
        max_workers: Optional[int] = None,
        max_keys: int = 1_000_000,
        max_dirs: int = 10,
        **storage_options,
    ):
        """
        Args:
            url: the url for which, ls is performed
            nested: if True, it attempts to make ls on each directory that it encounters. Otherwise, it will only yields
                the top-level paths and will not expand if the path is a directory
            max_workers (int): passed to the ThreadPoolExecutor. Only applicable if nested==True
            max_keys (int): maximum number of keys to keep in memory at the same time. If this number is reached, no new
                expansion on the currently discovered directories is done, until enough keys are yielded to make room
                for the new ones.
            max_dirs (int): maximum number of parallel ls operation.
            **storage_options (dict): kwargs to pass onto the fsspec filesystem initialization.
        """
        super().__init__()
        self.url = url
        self.protocol = get_protocol(url)
        self.nested = nested
        self.max_workers = max_workers
        self.max_keys = max_keys
        self.max_dirs = max_dirs
        self.storage_options = storage_options

    def __iter__(self) -> Iterator[str]:
        """Iterator that does ls and yield filepaths under the given url"""
        self.fs, _ = url_to_fs(self.url, **self.storage_options)
        urls = self.fs.ls(self.url, detail=False) if self.fs.exists(self.url) else []
        urls.sort()
        if self.nested:
            dirs = []
            with ThreadPoolExecutor(max_workers=self.max_workers) as pool:
                while len(urls) > 0 or dirs:
                    if len(urls) > 0:
                        url = urls.pop()
                        if self.fs.isdir(path=url):
                            future = AsyncContent(url, self.fs.ls, pool)

                            dirs.append(future)
                        else:
                            yield f"{self.protocol}{url}"
                    if (len(dirs) >= self.max_dirs and len(urls) < self.max_keys) or len(urls) == 0 and dirs:
                        d = dirs.pop(0).value()
                        urls.extend(d)

        else:
            for url in urls:
                yield f"{self.protocol}{url}"
